%
% $Id: lc_clean.sl,v 1.9 2003/06/20 14:23:12 dburke Exp $
%
% Script:
%   lc_clean.sl
%
% Aim:
%   CIAO version of Maxim Markevitch's lc_clean code
%   (see http://hea-www.harvard.edu/~maxim/axaf/acisbg/)
%   in S-Lang
%
% Thread:
%   http://cxc.harvard.edu/ciao/threads/acisbackground/
%
% CIAO version:
%   3.0
%
% Usage:
%  lc_clean( filename [, options array ] )
%
% where filename is the name of a file created by the CIAO tool
% dmextract and the optional options array is an associative
% array with the following keys (any others keys are ignored):
%
%   Keyword     Default
%               value  
%   -------------------
%   mean        NULL   
%   clip        3.0    
%   max_sigma   NULL     
%   max_scale   1.2      
%   outfile     NULL     
%   minfrac     0.1      
%   verbose     0        
%
% actually, the default value is taken from variables in the 'lc' namespace,
% where the variable name matches the keyword. So, if you only want to change
% clip for the current dataset you can send in the new value in an options 
% array (see the examples below); if you want to run the function many times
% with this different value, then it's probably easier to say
%  lc->clip = 4.0;
%
% Description:
%   filename must be the name of a file created by the CIAO dmextract tool.
% 
% Examples:
%   create a lightcurve of the background, with
%   a binning of 259.28 seconds. If it's called
%   lc.fits, then
%
%     unix% chips --slscript lc_clean.sl
%     chips> lc_clean( "lc.fits" )
%
%   or
%
%     chips> () = evalfile( "lc_clean.sl" )
%     chips> lc_clean( "lc.fits" )
%
%   To change options:
%
%     chips> opts = Assoc_Type []
%     chips> opts["mean"] = 0.665
%     chips> opts["outfile"] = "lc.gti"
%     chips> lc_clean( "lc.fits", opts )
%
% Notes:
%   The input lightcurve should have been created with a 
%   constant bin width (in seconds). The ACIS background datasets
%   were created with a bin-length of 259.28 seconds.
%
% To do:
%

implements("lc"); % define our own namespace

% set up the version string
variable version_str = "$Revision: 1.9 $";
version_str = version_str[[11:length(version_str)-4]];

% "global" variables
variable mean      = NULL;     % if defined use this as the mean level, otherwise ...
variable clip      = 3.0;      % use clip to calculate the mean background level

variable max_sigma = NULL;     % used to calculate the "good" background range from the mean ...
variable max_scale = 1.2;      % or this value is used

variable minfrac   = 0.1;      % minimum fraction of input #bins allowed in output
variable outfile   = NULL;     % name of output file

variable verbose   = 0;        % verbose level

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% local functions
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% output routines - thanks to John Davis
%
private define output (x) {
    () = fputs (x, stdout);
    () = fputs ("\n", stdout);
}
    
private define voutput () {
    variable args = __pop_args (_NARGS);
    output (sprintf (__push_args (args)));
}

private define vmsg () {
    variable args = __pop_args (_NARGS);

    % as cleared everything off the stack, we can return if verbose == 0
    variable verbose = args[0].value;
    if ( verbose == 0 ) return;

    output( sprintf( __push_args(args[[1:]]) ) );
}

private define myerror (estr) {
    error( "\n** ERROR **\n " + estr + "\n" );
}

% Usage:
%   x = [1:10];
%   stats = astats(x);
%
% Aim:
%   calculate a few "statistics" of the input array
%
%   Returned structure contains the following fields:
%       mean, min, max, npts
%
%   Currently for 1D data only
%
private define astats() {

    if ( _NARGS != 1 )
	usage( "( mean, min, max, npoints ) = astats(x);\n where x is a 1D array\n" );

    variable x = ();
    if ( typeof(x) != Array_Type )
        myerror( "Input variable to astats() should be an Array_Type, not " + 
                string(typeof(x)) );

    % should error on > 1D arrays

    % output stats - set up "defaults"
    variable stats = struct { mean, min, max, npts };
    set_struct_fields( stats, NULL, NULL, NULL, 0 );

    % handle empty arrays
    stats.npts = length(x);
    if ( stats.npts == 0 ) return stats;

    % set up the values using the first element
    stats.mean = 0.0;   % ensure that mean is a floating point
    stats.min  = x[0];
    stats.max  = x[0];

    variable value;
    foreach ( x ) {
        value = ();
        stats.mean += value; 
        if ( stats.min > value ) stats.min = value;
        if ( stats.max < value ) stats.max = value;
    }
    stats.mean = stats.mean / stats.npts;

    return stats;

} % astats()

% Usage:
%   ( hx. hy ) = histogram( x, xmin, xmax, xstep );
%
% Aim:
%   Create a histogram of the input array x (1D)
%   using bins centred at xmin to xmax, with width xstep
%
% Notes:
% . Given that S-Lang can create the bin centres pretty easily, 
%   perhaps we should just require the bin-centre array instead
%   of xmin, xmax, xstep -- or better, the low and high edges of 
%   each bin (inclusive/exclusive respectively)
%
% . Given that x = [0.0:10.0:1] misses out a bin centred on 10.
%   we ensure we have it by adding xstep onto xmax. 
%   Q: do we then have to filter out any values that are too large?
%

private define histogram () {

    variable usage_str = "( hx, hy ) = histogram(x,xmin,xmax,xstep);\n where x is a 1D array\n";
    if ( _NARGS != 4 ) usage( usage_str );

    variable x, xmin, xmax, xstep;
    ( x, xmin, xmax, xstep ) = ();
    % safety check
    if ( typeof(x) != Array_Type ) usage( usage_str );

    % create x axis - ie bin centres and edges
    variable hx    = [xmin:(xmax+xstep):xstep];
    variable hx_lo = hx - xstep/2;
    variable hx_hi = hx + xstep/2;

    variable nbins = length(hx);

    % bin the input array using these bins
    % note: we ignore those values outside the limits
    variable hy = Integer_Type [ nbins ];
    variable i;
    for ( i = 0; i < nbins; i++ ) {
	% how many bins lie within the current bin
        hy[i] = length(where( x >= hx_lo[i] and x < hx_hi[i] ) );  
    } % for: i
    
    return ( hx, hy );

} % histogram()

% qcurve( curve_num, chips_command )
%
% a simple wrapper around chips_eval for the curve command
%
private define qcurve (cnum,cstr) {
    () = chips_eval( "c " + string(cnum) + " " + cstr );
}

% qsym( curve_num, symbol_text )
%
% wrapper for ChIPS' symbol command: symbol_text can be
% an array of strings
%
private define qsym (cnum,text) {
    if ( typeof(text) == String_Type ) text = [ text ];
    foreach ( text ) {
        % don't seem to be able to say "symbol " + (), hence txt variable
        variable txt = ();
        qcurve( cnum, "symbol " + txt );
    }
}

% label_axes( xlabel, ylabel )
%
% label the current plot with the supplied x and y axes,
% and set the sizes to 1.2
%
private define label_axes ( xstr, ystr ) {
    () = chips_eval( "xlabel '" + xstr + "'" );
    () = chips_eval( "ylabel '" + ystr + "'" );
    () = chips_eval( "xlabel size 1.2" );
    () = chips_eval( "ylabel size 1.2" );
    () = chips_eval( "tickvals size 1.2" );

} % label_axes()

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% public routine
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

public define lc_clean () {

    %
    % handle the user input
    %
    variable usage_str = "lc_clean(filename[,opts_array])\n" + 
	" where filename is the name of the output of the dmextract tool\n" +
	" and opts_array is an optional associative array.";

    variable filename, useropts;

    switch ( _NARGS )
      {
	case 2 :
          ( filename, useropts ) = ();
          if ( typeof(useropts) != Assoc_Type ) usage( usage_str );
      }
      {
	case 1 :
          filename = ();
          useropts = NULL;
      }
      {
        usage( usage_str );
      }

    if ( typeof(filename) != String_Type ) usage( usage_str );

    % setup the default options - there are better ways to do this!
    variable mean      = lc->mean;
    variable clip      = lc->clip;
    variable max_scale = lc->max_scale;
    variable max_sigma = lc->max_sigma;
    variable minfrac   = lc->minfrac;
    variable outfile   = lc->outfile;
    variable verbose   = lc->verbose;

    % handle the user options - still not very clever
    if ( useropts != NULL ) {
        if ( assoc_key_exists( useropts, "mean" ) )      mean      = useropts["mean"];
        if ( assoc_key_exists( useropts, "clip" ) )      clip      = useropts["clip"];
        if ( assoc_key_exists( useropts, "max_scale" ) ) max_scale = useropts["max_scale"];
        if ( assoc_key_exists( useropts, "max_sigma" ) ) max_sigma = useropts["max_sigma"];
        if ( assoc_key_exists( useropts, "minfrac" ) )   minfrac   = useropts["minfrac"];
        if ( assoc_key_exists( useropts, "outfile" ) )   outfile   = useropts["outfile"];
        if ( assoc_key_exists( useropts, "verbose" ) )   verbose   = useropts["verbose"];
    } % if: useropts != NULL

    % let the user know what parameters are being used
    %
    vmsg( verbose, "Version of code: %s", lc->version_str );
    vmsg( verbose, "Parameters used to clean the lightcurve are:" );
    vmsg( verbose, "  mean      = %S", mean );
    vmsg( verbose, "  clip      = %S", clip );
    vmsg( verbose, "  max_scale = %S", max_scale );
    vmsg( verbose, "  max_sigma = %S", max_sigma );
    vmsg( verbose, "  minfrac   = %S", minfrac );
    vmsg( verbose, "  outfile   = %S", outfile );
    vmsg( verbose, "  verbose   = %S", verbose );
    vmsg( verbose, "" );

    % read in the lightcurve 
    % - using 'cols count_rate,exposure' filter should ensure they exist
    %
    variable lcurve = readfile( filename + "[cols time,exposure,count_rate]" );
    if ( lcurve == NULL )
	myerror( "Unable to read input lightcurve." );

    % safety checks
    variable ntot = lcurve._nrows;
    if ( ntot < 1 ) myerror( "No data read in from the lightcurve." );
    if ( ntot < 2 ) myerror( "Only 1 bin in lightcurve." );
    voutput( "Total number of bins in lightcurve   = %d", ntot );

    % what's the bin-size? 
    % we assume that the lightcurve has a constant bin-width;
    % however, GTI blocks may mean that we do not have continuous data,
    % so I don't know how that will work. So we find the max value,
    % and the number of bins with exposure times < this value.
    %
    variable delta_t = max( lcurve.EXPOSURE );
    voutput( "Max length of one bin                = %g s", delta_t );
    voutput( "Num. bins with a smaller exp. time   = %d", 
	length(where( lcurve.EXPOSURE != delta_t )) );

    %  we remove all elements that have a zero count_rate here
    %
    variable idx = where( lcurve.COUNT_RATE > 0 );
    if ( length(idx) == 0 ) 
	myerror( "Lightcurve does not contain any bins with count_rate > 0." );

    variable rate = lcurve.COUNT_RATE[idx];
    variable time = lcurve.EXPOSURE[idx];

    % useful to know if there are any "drop-out" bins in the lightcurve
    %
    variable num_time0 = ntot - length(rate);
    voutput( "Number of bins with a rate of 0 ct/s = %d\n", num_time0 );

    % calculate stats of count_rate column; useful even if fixing the mean value
    variable orig_stats = astats( rate );

    % calculate a histogram of the input count_rates (INcluding bins==0)
    %
    % the bin width of 0.05 counts/s should be okay for "reasonable"
    % light curves, although really it should be changed to match the input data
    % - we make it smaller if the range between min and max is < 0.25
    %
    % since we want to include rates of 0, have to fiddle the min level for the
    % calculation to be 0 if any such bins occur (orig_stats.min is guaranteed to
    % be > 0 since it's calculated from rate, which has been filtered of count_rate==0
    % bins
    %
    variable hist_rate, hist_num, r1;
    variable width = 0.05;
    if ( num_time0 > 0 )
        r1 = 0.0;
    else
        r1 = orig_stats.min;
    width = min( [ width, (orig_stats.max-r1)/5.0 ] );
    ( hist_rate, hist_num ) = histogram( lcurve.COUNT_RATE, r1, orig_stats.max, width );

    % start drawing things - note we only call "redraw on" IF there's
    % an error - ie so the user gets a plot of the lightcurve, which might
    % be enlightening - but we don't get flickering if there's no error
    %
    () = chips_eval( "redraw off" );
    () = chips_eval( "clear" );
    () = chips_eval( "split 2" );
    () = chips_eval( "split gap 0.1" );
 
    % protect any underscores in the filename
    %
    variable title;
    ( title, ) = strreplace( filename, "_", "\\_", strlen(filename) );
    () = chips_eval( "title 'Lightcurve: " + title + "'" );
    () = chips_eval( "title size 1.2" );

    % plot 1: lightcurve -- ALL points
    %
    () = chips_eval( "d 1" );
    () = curve( lcurve.TIME - lcurve.TIME[0], lcurve.COUNT_RATE );
    qsym( 1, [ "bigpoint", "size 1.2" ] );
    label_axes( "\\Delta Time (s)", "Background Rate (/s)" );

    % plot 2: histogram
    %
    () = chips_eval( "d 2" );
    () = curve( hist_rate, hist_num );
    qsym( 1, "none" );
    qcurve( 1, "step" );
    label_axes( "Background Rate (/s)", "Number" );

    % if the user has supplied a mean rate then we use that, otherwise
    % we have to calculate a "sigma-clipped" value
    %
    if ( mean == NULL ) {

        % - note that the sigma is calculated assuming "poisson" statistics
        %   for the mean value (delta_t is needed to account for scaling from number
        %   to a rate)
        %
        variable orig_sigma = sqrt( orig_stats.mean / delta_t );

        % what are the "sigma-clipped" limits?
        variable minval = orig_stats.mean - clip * orig_sigma;
        variable maxval = orig_stats.mean + clip * orig_sigma;

        % find mean of all filtered bins
	idx = where( rate > minval and rate < maxval );
        variable npts_filt = length(idx);
        if ( npts_filt == 0 ) {
            () = chips_eval( "redraw on" ); % show the current plot
	    myerror( "Unable to calculate mean: either exclude any flares or fix mean rate" );
        }

        variable filt_stats = astats( rate[idx] );
        mean = filt_stats.mean;

        voutput( "Calculated an initial mean (sigma-clipped) rate of %g ct/s", mean );

    } else {

        voutput( "Using a fixed mean rate of %g ct/s", mean );

    } % if mean != NULL

    % limits for lightcurve are calculated either as a straight factor
    % based on the mean level, or using "sigma clipping"
    %
    variable minlim, maxlim;
    if ( max_sigma != NULL ) {
      minlim = mean - max_sigma * sqrt( mean / delta_t );
      maxlim = mean + max_sigma * sqrt( mean / delta_t );
      vmessage( "Lightcurve limits clipped using %g sigma's about this mean", max_sigma );
    } else {
      minlim = mean / max_scale;
      maxlim = mean * max_scale;
      vmessage( "Lightcurve limits use a scale factor of %g about this mean", max_scale );
    }
    vmessage( "Filtering lightcurve between rates of %g and %g ct/s", minlim, maxlim );

    % find the acceptable times
    variable good_times = where( lcurve.COUNT_RATE > minlim and lcurve.COUNT_RATE < maxlim );
    variable ngood = length( good_times );
    vmessage( "Number of good time bins (drawn in green) = %d", ngood );
    if ( ngood == 0 ) {
        () = chips_eval( "redraw on" ); % show the current plot
        myerror( "Unable to find any bins within calculated rates" );
    }

    % finish off the plot by drawing on the "good" times in green
    %
    () = chips_eval( "redraw off" );

    % Plot 1: lightcurve -- GOOD times
    % mark on the allowed points as green, and make them a bit bigger
    %
    () = chips_eval( "d 1" );
    () = curve( lcurve.TIME[good_times] - lcurve.TIME[0], lcurve.COUNT_RATE[good_times] );
    qsym( 2, [ "bigpoint", "size 1.5", "green" ] );

    % Plot 2: histogram
    % draw, in green, those rates that we've selected
    %
    () = chips_eval( "d 2" );

    variable hist_rate_g, hist_num_g; 
    ( hist_rate_g, hist_num_g ) = 
	histogram( lcurve.COUNT_RATE[good_times], orig_stats.min, orig_stats.max, width );
    () = curve( hist_rate_g, hist_num_g );

    qsym( 2, "none" );
    qcurve( 2, "step" );
    qcurve( 2, "green" );

    () = chips_eval( "redraw on" );

    % what's the new mean rate?
    variable fstats = astats( lcurve.COUNT_RATE[good_times] );
    voutput( "Mean level of filtered lightcurve = %g ct/s", fstats.mean );

    % only do this check AFTER drawing selected bins
    if ( (ngood+0.0)/ntot < minfrac )
        myerror( "Fraction of bins that are good is below limit of " + string(minfrac) );

    % create the GTI file - if required
    %
    if ( outfile != NULL ) {
        output( "\nCreating GTI file" );

        variable retval = 0;

        % set the parameters
        retval |= system( "punlearn dmgti" );
        retval |= system( "pset dmgti infile=" + filename );
        retval |= system( "pset dmgti outfile=" + outfile );
        retval |= system( 
	    sprintf( "pset dmgti userlimit=\"(count_rate>%g && count_rate<%g)\"", minlim, maxlim )
		);
	if ( retval != 0 ) myerror( "Unable to set parameters for dmgti" );

	% run dmgti
        retval = system( "dmgti mode=h clobber=yes" );
	if ( retval != 0 ) myerror( "Unable to run dmgti" );

        voutput( "Created: %s", outfile );

    } % if: outfile != NULL

    return;

} % lc_clean()

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% end of code
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

