%
% $Id: paramest.sl,v 1.13 2005/11/14 16:38:57 dburke Exp $
%
% Summary:
%  Provide an interactive interface for running the parameter-estimation
%  routines.
%
% Requires:
%   CIAO >= 3.2
%
% Threads:
%  http://cxc.harvard.edu/sherpa/threads/confidence/
%  http://chandra.ledas.ac.uk/sherpa/threads/confidence/
%
% Usage:
%  See http://cxc.harvard.edu/ciao/download/scripts/ or
%  http://chandra.ledas.ac.uk/ciao/download/scripts/
%  for instructions on how to install S-Lang scripts and parameter
%  files.
%
%  Ensure that paramest.par is available (ie 'paccess paramest'
%  returns a file name) and then
%
%    sherpa> require("paramest")
%
%  or - in your ~/.sherparc (change the filename to match location)
%
%    require("paramest");
%
%  and then 
%
%   sherpa> rproj("clus.kt clus.abundanc")
%
%  or  
%
%   sherpa> rproj("clus.kt clus.abundanc","xnum=21")
%
% Documentation:
%  
%  - ahelp paramest
%  - ahelp <function name> (e.g. 'ahelp rproj')
%  - http://cxc.harvard.edu/ciao/ahelp/paramest.html
%  - http://chandra.ledas.ac.uk/ciao/ahelp/paramest.html
%
% Functions defined:
%   rproj( parlist [, option1, option2, ... ] )
%     Runs REGION-PROJECTION
%
%   runc( parlist [, option1, option2, ... ] )
%     Runs REGION-UNCERTAINTY
%
%   iproj( par [, option1, option2, ... ] )
%     Runs INTERVAL-PROJECTION
%
%   iunc( par [, option1, option2, ... ] )
%     Runs INTERVAL-UNCERTAINTY
%
%   proj( parlist [, option1, option2, ... ] )
%     Runs PROJECTION
%
%   unc( parlist [, option1, option2, ... ] )
%     Runs UNCERTAINTY
%
%   cov( parlist [, option ] )
%     Runs COVARIANCE [no need for multiple parameters here]
%
% Requires:
%   paramest.par in PFILES path (i.e. can plist/pset/punlearn paramest)
%
% Aim:
%   Provides an alternative interface to the Sherpa
%   parameter estimation commands like REGION-PROJECTION
%   and its S-Lang form run_regproj().
%
%   Here we use a parameter interface to set parameter values
%   - mainly so that we can let the parameter interface deal
%     with querying the values rather than dealing with I/O
%     ourselves
%   - as of v1.10 we only use the parameter interface to make
%     the I/O easy to write; the default values are taken from
%     the relevant Sherpa configuration variable
%     (so 'pset paramest foo=blah' is pointless)
%
%   The aim is to have a function that is easy to use for the
%   casual user: people using these functions many times will
%   likely find that this interface gets in the way, since it
%   always asks for parameters. Such users are suggested to
%   use a different interface.
%
% Notes:
%   Any changes to the parameter file made by the command-line
%   tools (pset et al) or their S-Lang equivalents  - 
%   import("paramio"); ... - have NO effect since we take the
%   default values from the state object. You can still set
%   parameter values when the function is called -
%     ie rproj("clus.kt clus.abundanc","xlog+")
%
%   If the user breaks out of the routine (control-c) then there is 
%   a possibility that strange things might happen with the parameter
%   interface when the routine is called again. We attempt to work
%   around the latter issue by having a "global" variable (at least 
%   global to this file/compilation unit) which contains the
%   Param_File_Type value for the last opened file. Then when a
%   function is called we check that this parameter is NULL and - 
%   if not - close it. This seems to work.
%
%   The use of the parameter library to access information is
%   not a necessity here; we use it since it provides an interface
%   that is familiar to users AND means we do not have to write
%   our own input/output routines for querying the user (although
%   these would not be too hard).
%
% To Do:
%   - on a 'usage error' list the parameters used by a function
%     so the user can find them out just by calling the function
%     with no parameters
%
%   - for the region-proj/unc routines, should we default to setting
%     chips.mingridsize equal to min(xnum,ynum) since this gives
%     (or at least seems to) nicer plots than the value of 50?
%     We would want a hidden parameter to control whether this was
%     used or not.
%
% Changes:
%   14 Nov 05 DJB v1.13
%   - work around a bug in the paccess() routine
%
%   02 Nov 04 DJB v1.12
%   - updated for CIAO 3.2:
%     * Updated rproj/unc() for fixes to run_regproj/unc()
%     * Updated iproj/unc() for fixes to run_intproj/unc()
%       - removes the error message seen in CIAO 3.1/3.0.2
%     * Use require rather than import for paramio module
%     * The script can be loaded via 'require("paramest");'
%
%   03 Dec 03 DJB v1.11
%   - updated for CIAO 3.0.2:
%     * run_regproj/unc() returns extra information that they are 
%       not supposed to
%     * run_intproj/unc() fail with a "struct has no field named config"
%       error message when successful. We trap this - using an ERROR_BLOCK
%       - so all the user sees is an extra error message go whizzing by
%     * removed the extra display of the parameter values in runc()
%     * some minor clean ups (mainly involving read/write of
%       boolean parameters).
%

require("paramio");

% We could only do this check the first time the code is
% evaluated but it is not very expensive and we do not expect
% the code to be re-evaluated often, so we do it everytime.
%
private variable PARNAME = "paramest";
private variable fp = NULL;
#iffalse  
if ( NULL == paccess(PARNAME) )
  verror( "\nError: unable to find the parameter file for %s\n",
          PARNAME );
#else
% In CIAO 3.2/3.3 paccess() does not return NULL on failure,
% as it should, so we use paramopen() instead.
%
fp = paramopen (PARNAME, "r");
if (NULL==fp)
  verror ("\nError: unable to find the parameter file for %s\n",
          PARNAME);
paramclose (fp);
fp = NULL; % improtant to clear fp
#endif

% utility routine to close the parameter file if it was open
%
private define _close_fp() {
  if ( NULL != fp ) {
    paramclose ( fp );
    fp = NULL;
  }
}

% It does not seem possible to clear the p_min/p_max fields
% of a parameter separately: the only way to do it is with a
% punlearn. 
%
% This used to be a lot more complex prior to v1.10 when
% we used to get the default values from the parameter file rather
% than the state object (and so we had to store the parameter
% values before calling punlearn and then reset them)
%
private define _reset_values() {

  % reset everything
  _close_fp();
  punlearn( PARNAME );

} % _reset_values()

%
% can we access the parameter file?
%
% This has not been updated to take advantage of the
% change introduced in CIAO 3.2, namely that the first
% argument of paramopen() can be set to NULL if the
% 3-argument form is used. In this situation there
% is no real advantage to changing the code
%
private define _open_paramfile (args) {

  % we do NOT want to call _reset_values() here
  _close_fp();
  if ( NULL == args ) fp = paramopen( PARNAME, "rwL" );
  else                fp = paramopen( PARNAME, "rwL", args );
  if ( NULL == fp )
    verror( "\nERROR: Unable to open the parameter file '%s'\n", PARNAME );
} % _open_paramfile

%
% handle the optional parameters
% - if optional arguments are given then we need to
%   create an argument array for paramopen() that
%   begins with the "tool" name.
%
private define _process_args () {
  !if ( _NARGS ) return NULL;

  variable args = String_Type [_NARGS+1];
  args[0] = PARNAME;

  variable i = 1;
  foreach ( __pop_args(_NARGS) ) {
    variable arg = ();
    args[i] = arg.value;
    i++;
  }

  return args;
} % _process_args()

%
% Sets the default values in the parameter file according to
% the values in the supplied structure:
%   obj can be a
%      sherpa_VisParEst_State
%      sherpa_Proj_State
%      sherpa_Cov_State
%      sherpa_Unc_State
%   structure
%
% We probably need more error checking in case the user has managed to
% trash the sherpa state object
%
private define _seed_paramfile( s ) {

  % reset everything (to get the min/max/prompt values)
  _reset_values();

  fp = paramopen( PARNAME, "rwLH" ); % mode is OTT?
  if ( NULL == fp ) verror( "\nERROR: Unable to open param file %s\n", PARNAME );

  switch ( typeof(s) )
  { case sherpa_VisParEst_State:
      %
      % fast             =  NULL / boolean
      % expfac           =  NULL / double
      % arange           =  boolean
      % min              =  double / double[2]
      % max              =  double / double[2]
      % log              =  integer / integer[2]
      % nloop            =  integer / integer[2]
      % sigma            =  double / double[]
      %
      if ( NULL != s.fast )   pputb( fp, "fastopt", s.fast );
      if ( NULL != s.expfac ) pputd( fp, "scale",   s.expfac );
      pputb( fp, "auto",     s.arange );
      if ( Array_Type == typeof(s.min) ) {
        pputd(   fp, "xmin", s.min[0] );
        pputd(   fp, "ymin", s.min[1] );
        pputd(   fp, "xmax", s.max[0] );
        pputd(   fp, "ymax", s.max[1] );
        pputb(   fp, "xlog", s.log[0] );
        pputb(   fp, "ylog", s.log[1] );
        pputi(   fp, "xnum", s.nloop[0] );
        pputi(   fp, "ynum", s.nloop[1] );
        pputstr( fp, "contours", strjoin(array_map( String_Type, &string, s.sigma), ",") );
      } else {
        pputd( fp, "xmin", s.min );
        pputd( fp, "xmax", s.max );
        pputb( fp, "xlog", s.log );
        pputi( fp, "xnum", s.nloop );
        pputd( fp, "sigma", s.sigma );
      }

  } % sherpa_VisParEst_State

  { case sherpa_Proj_State:
      %
      % fast             =  boolean
      % sigma            =  real
      %
      pputb( fp, "fastopt", s.fast );
      pputd( fp, "sigma",   s.sigma );

  } % sherpa_Proj_state
 
  { case sherpa_Cov_State:
      %
      % sigma            =  real
      %
      pputd( fp, "sigma",   s.sigma );

  } % sherpa_Cov_state
 
  { case sherpa_Unc_State:
      %
      % sigma            =  real
      % eps              =  real
      % remin            =  real
      %
      pputd( fp, "sigma",    s.sigma );
      pputd( fp, "tolsig",   s.eps );
      pputd( fp, "deltasig", s.remin );

  } % sherpa_Proj_state
 
  {
    paramclose(fp);
    fp = NULL;
    verror( "\nInternal error: %s sent a %S object\n",
            _function_name, typeof(s) );
  }

  paramclose(fp);
  fp = NULL;

} % _seed_paramfile()

%
% Currently a no-op: intended to allow the structure being edited to
% be restored to the state it was in prior to the function being called
% [should be controlled by a parameter value]
%
% low priority to implement
%
%%private define _restore_state() {}

% Utility function used when reading the
% min/max fields for an axis
%
% We let the parameter library check that the numeric values
% are within the hard limits of the parameter range
% - with a final check that min < max at the end
%
private define _get_minmax( coord, parname ) {

  variable par = get_par(parname);
  if ( NULL == par )
    verror( "\nERROR: Unable to get parameter information for %s\n",
            parname );

  % unfortunately we can not re-set the p_min/p_max fields
  % after setting them (or at least I do not know how)
  %
  variable label = coord + "min";
  pset( fp, label+".p_min", par.min );
  pset( fp, label+".p_max", par.max );
  pset( fp, label+".p_prompt", "Minimum value for " + par.name );
  variable minval = pgetd( fp, label );

  label = coord + "max";
  pset( fp, label+".p_min", par.min );
  pset( fp, label+".p_max", par.max );
  pset( fp, label+".p_prompt", "Maximum value for " + par.name );
  variable maxval = pgetd( fp, label );

  if ( maxval <= minval ) {
    %%_restore_state();
    _reset_values();
    verror( "\nERROR: min value must be less than max value for %s\n", parname );
  }

  return ( minval, maxval );

} % _get_minmax()

% converts a space or comma-separated list of names into
% an array of strings
%
% returns the array of parameter names
%
% - if the input string is empty AND the second parameter (npar)
%   is 0 then we assume the user wants to do all thawed parameters
%
private define _check_parlist() {

  variable parlist, npar;
  switch ( _NARGS )
  { case 2: ( parlist, npar ) = (); }
  { case 1: parlist = (); npar = 0; }
  { error( "Internal error: _check_parlist() called incorrectly" ); }

  % has the user specified any parameter names?
  variable parnames = strtok( strtrans( strtrim(parlist), ",", " " ) );
  variable nin = length(parnames);

  variable par, name, ctr;
  if ( 0 == nin ) {
    % expected a number of parameters
    if ( npar > 0 )
      verror( "\nERROR: unable to find any parameter names in input of '%s'\n",
              parlist );

    % no parameters supplied, so let us find all the thawed parameters
    variable allpars = get_par();

    nin = get_num_par_thawed();
    parnames = String_Type [nin];

    ctr = 0;
    foreach ( allpars ) {
      par = ();
      if ( par.frozen == 0 and "src" == par.type ) {
        parnames[ctr] = par.name;
        ctr++;
      }
    }
  }

  % do we have the correct number?
  if ( andelse { npar > 0 } { nin != npar } )
    verror( "\nERROR: expected %d parameter names but sent %d (%s)\n",
            npar, nin, parlist );

  % are the names valid?
  foreach ( parnames ) {
    name = ();
    par  = get_par( name );
    if ( NULL == par )
      verror( "\nERROR: parameter '%s' is not known\n", name );

    % check the argument is not frozen
    if ( par.frozen )
      verror( "\nERROR: parameter '%s' is frozen and can not be fit\n", name );

  } % foreach parnames

  return parnames;
} % _check_parlist

% print up useful information about the parameter(s)
% - how about including any error estimates?
%   would need to loop through get_proj, get_unc (or whatever
%   their names are) to get the values
%
private define _print_parinfo( parnames ) {
  message( "Parameter name:        Curr Val           Limits" );
  variable name, par;
  foreach ( parnames ) {
    name = ();
    par  = get_par( name );
    vmessage( "%-20s %10g  %10g %10g", name, par.value, par.min, par.max );
  }
  message( "" );
} % _print_parinfo

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Usage:
%   rproj( parlist )
%   rproj( parlist, val1, ..., valN )
%
% Aim:
%   Runs REGION-PROJECTION using the set-up described in the
%   'paramest' parameter file with the two parameters given in
%   parlist (space or comma-separated list)
%
%   The remaining parameters are read from the 'paramest'
%   parameter file. We use the following parameters
%   (not necessarily in this order):
%     auto       - are the limits calculated automatically?
%       auto=yes
%         scale  - factor to scale limits calculated by COVAR
%       auto=no
%         xmin   - minimum value for X parameter
%         xmax   - maximum value for X parameter
%         ymin   - minimum value for Y parameter
%         ymax   - maximum value for Y parameter
%     xlog       - should X axis be calculated using logarathmic spacing?
%     xnum       - number of points along X axis
%     ylog       - should X axis be calculated using logarathmic spacing?
%     ynum       - number of points along Y axis
%     contours   - comma-separated list of contour levels [units=sigma]
%     fastopt    - use LM or SIMPLEX as the optimiser instead of METHOD 
%
%   Note that the default values for the parameters are taken from
%   the values in the relevant sherpa.* configuration variable and
%   NOT the parameter file itself. This is a change introduced in
%   v1.10.
%
%   If more than two arguments are supplied these are
%   assumed to specify parameter values in the form
%   "name=value" - ie as a string. The name must be supplied
%
% Examples:
%   Run region-projection for the two parameters (clus.kt and gal.nh)
%   All values will be prompted for.
%
%     sherpa> rproj("clus.kt gal.nh")
%
%   Here the xnum parameter is set to 23 and it will NOT be
%   prompted for.
%
%     sherpa> rproj("clus.kt gal.nh","xnum=23")
%
define rproj() {

  if ( _NARGS < 1 )
    usage(
	"rproj( parlist [, arg1, ..., argN ] )\n\n" +
	"where parlist are the 2 parameter names (e.g. \"clus.kt gal.nh\")." 
         );

  % process the command-line
  variable alist = __pop_args(_NARGS-1);
  variable args = _process_args( __push_args(alist) );
  variable parlist  = ();

  % are the names valid?
  variable parnames = _check_parlist( parlist, 2 );
 
  % print up useful information about the parameters
  _print_parinfo( parnames );
  variable xname = parnames[0];
  variable yname = parnames[1];

  % seed the parameter file with the values from the
  % sherpa.regproj structure
  %
  _seed_paramfile( sherpa.regproj );

  % can we access the parameter file?
  _open_paramfile( args );

  % store the old state object settings
  % - not sure what to do about this at the moment
  %%variable oldvals = @sherpa.regproj;

  % get the settings from the parameter file
  %

  % A) limits
  variable auto = pgetb( fp, "auto" );
  if ( auto ) {
    sherpa.regproj.arange = 1;
    variable scale = pgetd( fp, "scale" );
    sherpa.regproj.expfac = scale;

  } else {
    sherpa.regproj.arange = 0;

    variable xmin, xmax;
    ( xmin, xmax ) = _get_minmax( "x", xname );

    variable ymin, ymax;
    ( ymin, ymax ) = _get_minmax( "y", yname );

    sherpa.regproj.min = [ xmin, ymin ];
    sherpa.regproj.max = [ xmax, ymax ];
  }

  % B) spacing
  variable xlog = pgetb( fp, "xlog" );
  variable xnum = pgeti( fp, "xnum" );

  variable ylog = pgetb( fp, "ylog" );
  variable ynum = pgeti( fp, "ynum" );

  sherpa.regproj.log   = [ xlog, ylog ];
  sherpa.regproj.nloop = [ xnum, ynum ];

  % C) contour levels
  % - minimal error checking
  % - we allow the numbers to be separated by spaces or commas
  %
  variable parval = pget( fp, "contours" );
  variable sigma = array_map( Double_Type, &atof, strtok( parval, ", " ) );
  if ( length(sigma) < 1 ) {
    %%sherpa.regproj = oldvals;
    _reset_values();
    error( "\nERROR: must be at least one contour level\n" );
  }
  if ( sum(sigma<=0.0) > 0 ) {
    %%sherpa.regproj = oldvals;
    _reset_values();
    error( "\nERROR: contours parameter must contain positive values\n" );
  }
  sherpa.regproj.sigma = sigma;

  % D) optimisation setting
  sherpa.regproj.fast = pgetb( fp, "fastopt" );

  % we have finished with the parameter file
  _close_fp();

  % run the projection
  variable retval = run_regproj(parnames);
  if ( NULL == retval ) {
    %%sherpa.regproj = oldvals;
    verror( "\nERROR: Unable to run REGION-PROJECTION on %s and %s\n", xname, yname );
  }

  % Informational message
  vmessage( "\nThe REGION-PROJECTION run has finished for %s and %s.", xname, yname );
  message( "The get_regproj() command can be used to access the surface data" );
  message( "of this run.\n" );

  % restore the old settings
  %%sherpa.regproj = oldvals;

} % rproj()

% Usage:
%   runc( parlist )
%   runc( parlist, val1, ..., valN )
%
% Aim:
%   Runs REGION-UNCERTAINTY using the set-up described in the
%   'paramest' parameter file with the two parameters given in
%   parlist (space or comma-separated list)
%
%   The remaining parameters are read from the 'paramest'
%   parameter file. We use the following parameters
%   (not necessarily in this order):
%     auto       - are the limits calculated automatically?
%       auto=yes
%         scale  - factor to scale limits calculated by COVAR
%       auto=no
%         xmin   - minimum value for X parameter
%         xmax   - maximum value for X parameter
%         ymin   - minimum value for Y parameter
%         ymax   - maximum value for Y parameter
%     xlog       - should X axis be calculated using logarathmic spacing?
%     xnum       - number of points along X axis
%     ylog       - should X axis be calculated using logarathmic spacing?
%     ynum       - number of points along Y axis
%     contours   - comma-separated list of contour levels [units=sigma]
%
%   Note that the default values for the parameters are taken from
%   the values in the relevant sherpa.* configuration variable and
%   NOT the parameter file itself. This is a change introduced in
%   v1.10.
%
%   If more than two arguments are supplied these are
%   assumed to specify parameter values in the form
%   "name=value" - ie as a string. The name must be supplied
%
% Examples:
%   Run region-uncertainty for the two parameters (clus.kt and gal.nh)
%   All values will be prompted for.
%
%     sherpa> runc("clus.kt gal.nh")
%
%   Here the xnum parameter is set to 23 and it will NOT be
%   prompted for.
%
%     sherpa> runc("clus.kt gal.nh","xnum=23")
%
define runc() {

  if ( _NARGS < 1 )
    usage(
	"runc( parlist [, arg1, ..., argN ] )\n\n" +
	"where parlist are the 2 parameter names (e.g. \"clus.kt gal.nh\")." 
         );

  % process the command-line
  variable alist = __pop_args(_NARGS-1);
  variable args = _process_args( __push_args(alist) );
  variable parlist  = ();

  % are the names valid?
  variable parnames = _check_parlist( parlist, 2 );
 
  % print up useful information about the parameters
  _print_parinfo( parnames );
  variable xname = parnames[0];
  variable yname = parnames[1];

  % seed the parameter file with the values from the
  % sherpa.regunc structure
  %
  _seed_paramfile( sherpa.regunc );

  % can we access the parameter file?
  _open_paramfile( args );

  % store the old state object settings
  % - not sure what to do about this at the moment
  %%variable oldvals = @sherpa.regunc;

  % get the settings from the parameter file
  %

  % A) limits
  variable auto = pgetb( fp, "auto" );
  if ( auto ) {
    sherpa.regunc.arange = 1;
    variable scale = pgetd( fp, "scale" );
    sherpa.regunc.expfac = scale;

  } else {
    sherpa.regunc.arange = 0;

    variable xmin, xmax;
    ( xmin, xmax ) = _get_minmax( "x", xname );

    variable ymin, ymax;
    ( ymin, ymax ) = _get_minmax( "y", yname );

    sherpa.regunc.min = [ xmin, ymin ];
    sherpa.regunc.max = [ xmax, ymax ];
  }

  % B) spacing
  variable xlog = pgetb( fp, "xlog" );
  variable xnum = pgeti( fp, "xnum" );

  variable ylog = pgetb( fp, "ylog" );
  variable ynum = pgeti( fp, "ynum" );

  sherpa.regunc.log   = [ xlog, ylog ];
  sherpa.regunc.nloop = [ xnum, ynum ];

  % C) contour levels
  % - minimal error checking
  % - we allow the numbers to be separated by spaces or commas
  %
  variable parval = pget( fp, "contours" );
  variable sigma = array_map( Double_Type, &atof, strtok( parval, ", " ) );
  if ( length(sigma) < 1 ) {
    %%sherpa.regunc = oldvals;
    _reset_values();
    error( "\nERROR: must be at least one contour level\n" );
  }
  if ( sum(sigma<=0.0) > 0 ) {
    %%sherpa.regunc = oldvals;
    _reset_values();
    error( "\nERROR: contours parameter must contain positive values\n" );
  }
  sherpa.regunc.sigma = sigma;

  % we have finished with the parameter file
  _close_fp();

  % run the uncertainty command
  variable retval = run_regunc(parnames);
  if ( NULL == retval ) {
    %%sherpa.regunc = oldvals;
    verror( "\nERROR: Unable to run REGION-UNCERTAINTY on %s and %s\n", xname, yname );
  }

  % Informational message
  vmessage( "\nThe REGION-UNCERTAINTY run has finished for %s and %s.", xname, yname );
  message( "The get_regunc() command can be used to access the surface data" );
  message( "of this run.\n" );

  % restore the old settings
  %%sherpa.regunc = oldvals;

} % runc()

% Usage:
%   iproj( name )
%   iproj( name, val1, ..., valN )
%
% Aim:
%   Runs INTERVAL-PROJECTION using the set-up described in the
%   'paramest' parameter file with the parameter being
%   set by name (a string).
%
%   The remaining parameters are read from the 'paramest'
%   parameter file. We use the following parameters
%   (not necessarily in this order):
%     auto       - are the limits calculated automatically?
%       auto=yes
%         scale  - factor to scale limits calculated by COVAR
%         sigma  - Number of sigma
%       auto=no
%         xmin   - minimum value for parameter
%         xmax   - maximum value for parameter
%     xlog       - should axis be calculated using logarathmic spacing?
%     xnum       - number of points along axis
%     fastopt    - use LM or SIMPLEX as the optimiser instead of METHOD 
%
%   Note that the default values for the parameters are taken from
%   the values in the relevant sherpa.* configuration variable and
%   NOT the parameter file itself. This is a change introduced in
%   v1.10.
%
%   If more than one argument is supplied they are
%   assumed to specify parameter values in the form
%   "name=value" - ie as a string. The name must be supplied
%
% Examples:
%   Run interval-projection for the parameter clus.kt
%   All values will be prompted for.
%
%     sherpa> iproj("clus.kt")
%
%   Here the xnum parameter is set to 23 and it will NOT be
%   prompted for.
%
%     sherpa> iproj("clus.kt","xnum=23")
%
define iproj() {

  if ( _NARGS < 1 )
    usage(
	"iproj( name [, arg1, ..., argN ] )\n\n" +
	"where name is the parameter name (e.g. \"clus.kt\")." 
         );

  % process the command-line
  variable alist = __pop_args(_NARGS-1);
  variable args = _process_args( __push_args(alist) );
  variable xname = ();

  % is the name valid?
  () = _check_parlist( xname, 1 );

  % print up useful information about the parameter
  _print_parinfo( [ xname ] );

  % seed the parameter file with the values from the
  % sherpa.intproj structure
  %
  _seed_paramfile( sherpa.intproj );

  % can we access the parameter file?
  _open_paramfile( args );

  % store the old state object settings
  % - not sure what to do about this at the moment
  %%variable oldvals = @sherpa.regproj;

  % get the settings from the parameter file
  %

  % A) limits
  variable auto = pgetb( fp, "auto" );
  if ( auto ) {
    sherpa.intproj.arange = 1;
    variable scale = pgetd( fp, "scale" );
    sherpa.intproj.expfac = scale;

    % sigma value
    % - minimal error checking
    %
    variable sigma = pgetd( fp, "sigma" );
    sherpa.intproj.sigma = sigma;

  } else {
    sherpa.intproj.arange = 0;

    variable xmin, xmax;
    ( xmin, xmax ) = _get_minmax( "x", xname );

    sherpa.intproj.min = xmin;
    sherpa.intproj.max = xmax;
  }

  % B) spacing
  variable xlog = pgetb( fp, "xlog" );
  variable xnum = pgeti( fp, "xnum" );

  sherpa.intproj.log   = xlog;
  sherpa.intproj.nloop = xnum;

  % D) optimisation setting
  sherpa.intproj.fast = pgetb( fp, "fastopt" );

  % we have finished with the parameter file
  _close_fp();

  % run the projection
  %
  variable retval = run_intproj(xname);
  if ( NULL == retval ) {
    %%sherpa.regproj = oldvals;
    verror( "\nERROR: Unable to run INTERVAL-PROJECTION on %s\n", xname );
  }

  % Informational message
  vmessage( "\nThe INTERVAL-PROJECTION run has finished for %s.", xname );
  message( "The get_intproj() command can be used to access the data" );
  message( "of this run.\n" );

  % restore the old settings
  %%sherpa.regproj = oldvals;

} % iproj()

% Usage:
%   iunc( name )
%   iunc( name, val1, ..., valN )
%
% Aim:
%   Runs INTERVAL-UNCERTAINTY using the set-up described in the
%   'paramest' parameter file with the parameter being
%   set by name (a string).
%
%   The remaining parameters are read from the 'paramest'
%   parameter file. We use the following parameters
%   (not necessarily in this order):
%     auto       - are the limits calculated automatically?
%       auto=yes
%         scale  - factor to scale limits calculated by COVAR
%         sigma  - Number of sigma
%       auto=no
%         xmin   - minimum value for parameter
%         xmax   - maximum value for parameter
%     xlog       - should axis be calculated using logarathmic spacing?
%     xnum       - number of points along axis
%
%   Note that the default values for the parameters are taken from
%   the values in the relevant sherpa.* configuration variable and
%   NOT the parameter file itself. This is a change introduced in
%   v1.10.
%
%   If more than one argument is supplied they are
%   assumed to specify parameter values in the form
%   "name=value" - ie as a string. The name must be supplied
%
% Examples:
%   Run interval-uncertainty for the parameter clus.kt
%   All values will be prompted for.
%
%     sherpa> iunc("clus.kt")
%
%   Here the xnum parameter is set to 23 and it will NOT be
%   prompted for.
%
%     sherpa> iunc("clus.kt","xnum=23")
%
define iunc() {

  if ( _NARGS < 1 )
    usage(
	"iunc( name [, arg1, ..., argN ] )\n\n" +
	"where name is the parameter name (e.g. \"clus.kt\")." 
         );

  % process the command-line
  variable alist = __pop_args(_NARGS-1);
  variable args = _process_args( __push_args(alist) );
  variable xname = ();

  % is the name valid?
  () = _check_parlist( xname, 1 );

  % print up useful information about the parameter
  _print_parinfo( [ xname ] );

  % seed the parameter file with the values from the
  % sherpa.intunc structure
  %
  _seed_paramfile( sherpa.intunc );

  % can we access the parameter file?
  _open_paramfile( args );

  % store the old state object settings
  % - not sure what to do about this at the moment
  %%variable oldvals = @sherpa.regproj;

  % get the settings from the parameter file
  %

  % A) limits
  variable auto = pgetb( fp, "auto" );
  if ( auto ) {
    sherpa.intunc.arange = 1;

    % sigma value
    % - minimal error checking
    %
    variable sigma = pgetd( fp, "sigma" );
    sherpa.intunc.sigma = sigma;

  } else {
    sherpa.intunc.arange = 0;

    variable xmin, xmax;
    ( xmin, xmax ) = _get_minmax( "x", xname );

    sherpa.intunc.min = xmin;
    sherpa.intunc.max = xmax;

  }

  % B) spacing
  variable xlog = pgetb( fp, "xlog" );
  variable xnum = pgeti( fp, "xnum" );

  sherpa.intunc.log   = xlog;
  sherpa.intunc.nloop = xnum;

  % we have finished with the parameter file
  _close_fp();

  % run the uncertainty command
  %
  variable retval = run_intunc(xname);
  if ( NULL == retval ) {
    %%sherpa.regproj = oldvals;
    verror( "\nERROR: Unable to run INTERVAL-UNCERTAINTY on %s\n", xname );
  }

  % Informational message
  vmessage( "\nThe INTERVAL-UNCERTAINTY run has finished for %s.", xname );
  message( "The get_intunc() command can be used to access the data" );
  message( "of this run.\n" );

  % restore the old settings
  %%sherpa.regproj = oldvals;

} % iunc()

% Usage:
%   proj( parlist )
%   proj( parlist, val1, ..., valN )
%
% Aim:
%   Runs PROJECTION using the set-up described in the
%   'paramest' parameter file with the parameter(s) being
%   set by parlist (a space or comma-separtated string
%   of parameters, can be "" for all thawed parameters).
%
%   The remaining parameters are read from the 'paramest'
%   parameter file. We use the following parameters
%   (not necessarily in this order):
%     sigma      - Number of sigma
%     fastopt    - use LM or SIMPLEX as the optimiser instead of METHOD 
%
%   Note that the default values for the parameters are taken from
%   the values in the relevant sherpa.* configuration variable and
%   NOT the parameter file itself. This is a change introduced in
%   v1.10.
%
%   If more than one argument is supplied they are
%   assumed to specify parameter values in the form
%   "name=value" - ie as a string. The name must be supplied
%
% Examples:
%   Run projection for the parameter clus.kt
%   All values will be prompted for.
%
%     sherpa> proj("clus.kt")
%
%   Either command runs projection for the parameters clus.kt and
%   clus.abundanc
%   All values will be prompted for.
%
%     sherpa> proj("clus.kt clus.abundanc")
%
%   Here the sigma parameter is set to 1.6 (ie 90%) and it will NOT be
%   prompted for.
%
%     sherpa> proj("clus.kt","sigma=1.6")
%
%   Here the default value for the sigma parameter is set to 1.6
%   but it will still be prompted for.
%
%     sherpa> sherpa.proj.sigma = 1.6
%     sherpa> proj("clus.kt")
%
define proj() {

  if ( _NARGS < 1 )
    usage(
	"proj( parlist [, arg1, ..., argN ] )\n\n" +
	"where parlist is a space or comma-separated string of parameter names (e.g. \"clus.kt\")." 
         );

  % process the command-line
  variable alist = __pop_args(_NARGS-1);
  variable args = _process_args( __push_args(alist) );
  variable parlist  = ();

  % are the names valid?
  variable parnames = _check_parlist(parlist);
  parlist = strjoin(parnames," ");

  % print up useful information about the parameter(s)
  _print_parinfo( parnames );

  % seed the parameter file with the values from the
  % sherpa.proj structure
  %
  _seed_paramfile( sherpa.proj );

  % can we access the parameter file?
  _open_paramfile( args );

  % store the old state object settings
  % - not sure what to do about this at the moment
  %%variable oldvals = @sherpa.regproj;

  % get the settings from the parameter file
  %

  % A) sigma value
  % - minimal error checking
  %
  variable sigma = pgetd( fp, "sigma" );
  sherpa.proj.sigma = sigma;

  % B) optimisation setting
  sherpa.proj.fast = pgetb( fp, "fastopt" );

  % we have finished with the parameter file
  _close_fp();

  % run the projection
  variable retval = run_proj(parnames);
  if ( NULL == retval ) {
    %%sherpa.regproj = oldvals;
    verror( "\nERROR: Unable to run PROJECTION on %s\n", parlist );
  }

  % Informational message
  vmessage( "\nThe PROJECTION run has finished for %s.", parlist );
  message( "The get_proj() command can be used to access the data" );
  message( "of this run.\n" );

  % restore the old settings
  %%sherpa.regproj = oldvals;

} % proj()

% Usage:
%   unc( parlist )
%   unc( parlist, val1, ..., valN )
%
% Aim:
%   Runs UNCERTAINTY using the set-up described in the
%   'paramest' parameter file with the parameter(s) being
%   set by parlist (a space or comma-separtated string
%   of parameters, can be "" for all thawed parameters).
%
%   The remaining parameters are read from the 'paramest'
%   parameter file. We use the following parameters
%   (not necessarily in this order):
%     sigma      - Number of sigma
%     toleps     - Tolerance for sigma (hidden)
%     deltasig   - Threshold for Stat value to change before refir (hidden)
%
%   Note that the default values for the parameters are taken from
%   the values in the relevant sherpa.* configuration variable and
%   NOT the parameter file itself. This is a change introduced in
%   v1.10.
%
%   If more than one argument is supplied they are
%   assumed to specify parameter values in the form
%   "name=value" - ie as a string. The name must be supplied
%
% Examples:
%   Run uncertainty for the parameter clus.kt
%   All values will be prompted for.
%
%     sherpa> unc("clus.kt")
%
%   Here the sigma parameter is set to 1.6 (ie 90%) and it will NOT be
%   prompted for.
%
%     sherpa> unc("clus.kt","sigma=1.6")
%
%   Here the default value for the sigma parameter is set to 1.6
%   but it will still be prompted for.
%
%     sherpa> sherpa.unc.sigma = 1.6
%     sherpa> unc("clus.kt")
%
define unc() {

  if ( _NARGS < 1 )
    usage(
	"unc( parlist [, arg1, ..., argN ] )\n\n" +
	"where parlist is a space or comma-separated string of parameter names (e.g. \"clus.kt\")." 
         );

  % process the command-line
  variable alist = __pop_args(_NARGS-1);
  variable args = _process_args( __push_args(alist) );
  variable parlist  = ();

  % are the names valid?
  variable parnames = _check_parlist(parlist);
  parlist = strjoin(parnames," ");

  % print up useful information about the parameter(s)
  _print_parinfo( parnames );

  % seed the parameter file with the values from the
  % sherpa.unc structure
  %
  _seed_paramfile( sherpa.unc );

  % can we access the parameter file?
  _open_paramfile( args );

  % store the old state object settings
  % - not sure what to do about this at the moment
  %%variable oldvals = @sherpa.regproj;

  % get the settings from the parameter file
  %

  % A) sigma value
  % - minimal error checking
  %
  sherpa.unc.sigma = pgetd( fp, "sigma" );

  % B) sigma tolerance
  sherpa.unc.eps = pgetd( fp, "tolsig" );

  % C) threshold change in Statistic to force a re-fit
  sherpa.unc.remin = pgetd( fp, "deltasig" );

  % we have finished with the parameter file
  _close_fp();

  % run the uncertainty
  variable retval = run_unc(parnames);
  if ( NULL == retval ) {
    %%sherpa.regproj = oldvals;
    verror( "\nERROR: Unable to run UNCERTAINTY on %s\n", parlist );
  }

  % Informational message
  vmessage( "\nThe UNCERTAINTY run has finished for %s.", parlist );
  message( "The get_unc() command can be used to access the data" );
  message( "of this run.\n" );

  % restore the old settings
  %%sherpa.regproj = oldvals;

} % unc()

% Usage:
%   cov( parlist )
%   cov( parlist, val1, ..., valN )
%
% Aim:
%   Runs COVARIANCE using the set-up described in the
%   'paramest' parameter file with the parameter(s) being
%   set by parlist (a space or comma-separtated string
%   of parameters, can be "" for all thawed parameters).
%
%   The remaining parameters are read from the 'paramest'
%   parameter file. We use the following parameter:
%     sigma      - Number of sigma
%
%   Note that the default values for the parameters are taken from
%   the values in the relevant sherpa.* configuration variable and
%   NOT the parameter file itself. This is a change introduced in
%   v1.10.
%
%   If more than one argument is supplied they are
%   assumed to specify parameter values in the form
%   "name=value" - ie as a string. This obviously does not have much meaning
%   for this function.
%   The name must be supplied
%
% Examples:
%   Run covariance for the parameter clus.kt
%   All values will be prompted for.
%
%     sherpa> cov("clus.kt")
%
%   Here the sigma parameter is set to 1.6 (ie 90%) and it will NOT be
%   prompted for.
%
%     sherpa> cov("clus.kt","sigma=1.6")
%
%   Here the default value for the sigma parameter is set to 1.6
%   but it will still be prompted for.
%
%     sherpa> sherpa.cov.sigma = 1.6
%     sherpa> cov("clus.kt")
%
define cov() {

  if ( _NARGS < 1 )
    usage(
	"cov( parlist [, arg1, ..., argN ] )\n\n" +
	"where parlist is a space or comma-separated string of parameter names (e.g. \"clus.kt\")." 
         );

  % process the command-line
  variable alist = __pop_args(_NARGS-1);
  variable args = _process_args( __push_args(alist) );
  variable parlist  = ();

  % are the names valid?
  variable parnames = _check_parlist(parlist);
  parlist = strjoin(parnames," ");

  % print up useful information about the parameter(s)
  _print_parinfo( parnames );

  % seed the parameter file with the values from the
  % sherpa.cov structure
  %
  _seed_paramfile( sherpa.cov );

  % can we access the parameter file?
  _open_paramfile( args );

  % store the old state object settings
  % - not sure what to do about this at the moment
  %%variable oldvals = @sherpa.regproj;

  % get the settings from the parameter file
  %

  % A) sigma value
  % - minimal error checking
  %
  sherpa.cov.sigma = pgetd( fp, "sigma" );

  % we have finished with the parameter file
  _close_fp();

  % run the covariance routine
  variable retval = run_cov(parnames);
  if ( NULL == retval ) {
    %%sherpa.regproj = oldvals;
    verror( "\nERROR: Unable to run COVARIANCE on %s\n", parlist );
  }

  % Informational message
  vmessage( "\nThe COVARIANCE run has finished for %s.", parlist );
  message( "The get_cov() command can be used to access the data" );
  message( "of this run.\n" );

  % restore the old settings
  %%sherpa.regproj = oldvals;

} % cov()

provide( "paramest" );

%%%
%%% End
%%%
