%
% $Id: regions.sl,v 1.2 2003/07/17 18:48:39 dburke Exp $
%
% Usage:
%   display_fef_regions( filename )
%
%   where filename includes any necessary spatial filtering
%  
% Example:
%   chips> () = evalfile( "regions.sl" );
%   chips> display_fef_regions( "evt2.fits[sky=region(src.reg)]" );
%
% or
%
%   unix% chips --slscript regions.sl
%   chips> display_fef_regions( "evt2.fits[sky=region(src.reg)]" );
%
% Thread:
%   http://cxc.harvard.edu/ciao/threads/show_fef_regions/
%
% NOTES:
%   There have been minimal changes made to this script to
%   take advantage of the new facilities in CIAO 3.0
%   (e.g. the caldb interface)
%
% Aim:
%   Display the CHIP coordinates for a region, and
%   indicate the calibration (FEF) regions - along
%   with the number of photons that lie within these regions
%
%   For ACIS data only. The file (or virtual file) supplied to the
%   routine *must* consist of data for only ONE ACIS chip, and 
%   *must* contain a TIME column (this is used to find out the
%   Focal Plane temperature of the observation). Now that Varmm
%   reads in the FITS header we *could* use that, but we don't.
%
%   A plot is created, showing the position of the photons in
%   the chip coordinate system, together with outlines showing the
%   borderrs of the FEF regions. Information on the number of 
%   photons falling within each region -- together with the
%   chip coordinates of the region -- are printed to the
%   screen.
%
%   The grids drawn on the screen may be off by a pixel, but
%   the values printed to the screen (and used to calculate
%   the number of photons within each region) should be correct.
%

% outupt routines - thanks to John Davis
%
define output (x) {
    () = fputs (x, stdout);
    () = fputs ("\n", stdout);
}
    
define voutput () {
    variable args = __pop_args (_NARGS);
    output (sprintf (__push_args (args)));
}

% return the min and max values of an array
%
define minmax() { % array

    if ( _NARGS != 1 )
	usage( "(min,max) = minmax(a)\nwhere a is an array of numbers.");

    variable a = ();
    if ( typeof(a) != Array_Type )
	error( "Error: input variable to minmax() should be an Array_Type, not %s",
		string(typeof(a)) );

    variable min, max, value;
    min = a[0];
    max = a[0];
    foreach (a) {
        value = ();
	if ( min > value ) min = value;
	if ( max < value ) max = value;
    }

    return ( min, max );

} % minmax()

%
% find the FP temperature for this observation
% and return an id into the region-spacing array
%
define find_temperature() { % tstart (seconds)

    if ( _NARGS != 1 )
        usage( "(temp,temp_id) = find_temperature( tstart )\nwhere tstart is the start of the observation (in s)." );

    variable tstart = ();
    variable temperature, id;

    % simple checks
    if ( tstart >= 65563264 )      { temperature = -120; id = 3; }
    else if ( tstart >= 53827200 ) { temperature = -110; id = 2; }
    else if ( tstart >= 50889600 ) { temperature = -100; id = 1; }
    else if ( tstart >= 48988800 ) { temperature =  -90; id = 0; }
    else {
	voutput( "Error: TSTART (%f) is too small.", tstart );
	output(  "       Unable to find temperature of Focal Plane." );
	return ( NULL, NULL );
    }

    return ( temperature, id );

} % find_temperature()

%
% Store the spacing of the FEF regions: they depend on
% chip temperature and chip ID
%
private variable nchips = 10;
private variable ntemp  = 4;
private variable DDX = Integer_Type[ntemp,nchips];
private variable DDY = Integer_Type[ntemp,nchips];

%    ACIS   0    1    2    3    4    5    6    7    8    9
DDX[0,*] = [ 256, 256, 256, 256, 256,  64, 256,  32, 256, 256 ];  %  -90
DDX[1,*] = [ 256, 256, 256, 256, 256,  64, 256,  32, 256, 256 ];  % -100
DDX[2,*] = [ 256, 256, 256, 256, 256,  64, 256,  32, 256, 256 ];  % -110
DDX[3,*] = [ 256, 256, 256, 256, 256,  64, 256,  32, 256, 256 ];  % -120

DDY[0,*] = [ 1024, 1024, 1024, 1024, 1024,   64, 1024,   32, 1024, 1024 ];
DDY[1,*] = [ 1024, 1024, 1024, 1024, 1024,   64, 1024,   32, 1024, 1024 ];
DDY[2,*] = [   32,   32,   32,   32,   32,   64,   32,   32,   32,   32 ];
DDY[3,*] = [   32,   32,   32,   32,   32,   64,   32,   32,   32,   32 ];

%
% We use a simple linked list to store the information
% on number of photons within a region (together with the
% region description).
%
% head = slist_insert( head, structure )
%
% inserts a structure (which should contain the elements
% in the commented Region_Type structure below) into the
% list whose first element is stored in head. The insertion
% is done so that the first element has the largest number of
% photons, the last element the smallest. The new head is
% returned after the insertion.
%
% typedef struct {
%    photons,    % number of photons
%    fraction,   % fraction of total photon distribution
%    region,     % string containing the chip region
%    next,       % reference to next node
%    prev        % reference to previous node
%} Region_Type;
%
% Would it have been easier to work from the bottom of
% the list up - ie avoid the need for both .next and .prev ?
%

define slist_insert() { % slist, structure (matching Region_Type above)

    if ( _NARGS != 2 )
        usage( "slist = slist_insert(slist,structure)" );

    variable slist, structure;
    ( slist, structure ) = ();

    % insert into slist
    %
    % If slist is empty, this is easy
    if ( slist == NULL ) {
        slist = structure;
	structure.next = NULL;
	structure.prev = NULL;

    } else {
	% loop through slist, looking either for the end of the list,
	% or the first structure whose photons field is smaller than
	% structure.photons
	%
	variable this = slist;
	variable smaller = NULL;
	variable prev;
	
	% separate out the searching from the insertion code
	% to improve readability
	while ( this != NULL and smaller == NULL ) {
	    if ( this.photons < structure.photons ) {
	        smaller = this;
	    } else {
		prev = this;
		this = this.next;
	    }
	} 

	if ( smaller != NULL ) {
	    structure.next = smaller;
	    structure.prev = smaller.prev;
	    smaller.prev   = structure;

	    if ( structure.prev == NULL ) {
		% inserting a new start-of-list element 
		slist = structure; 
	    } else {
		% inserting somewhere within the list
		structure.prev.next = structure;
	    }

	} else {
	    % we are at the end of the list
	    structure.prev = prev;
	    structure.next = NULL;
	    prev.next      = structure;
	}

    }

    return slist;

} % slist_insert()

% draws grid, and reports stats on those regions with counts
%
% the x and y arrays are needed so that the routine can calculate
% the number of photons within each region (and restrict the 
% calculation to just the relevant FEF regions).
%
define draw_grid() { % ccd_id, temperature_id, x, y

    if ( _NARGS != 4 )
	usage( "draw_grid(ccd_id,temperature_id,chipx,chipy)" );

    variable ccd_id, temp_id, x, y;
    ( ccd_id, temp_id, x, y ) = ();

    % checks - should check that x and y are arrays & the same size
    if ( ccd_id < 0 or ccd_id >= nchips )
	error( "Error: ccd_id must be in range 0 - %d, not %d.",
		(nchips-1), ccd_id );

    if ( temp_id < 0 or temp_id >= ntemp )
	error( "Error: temperature_id must be in range 0 - %d, not %d.",
		(ntemp-1), temp_id );

    % size of grid
    variable dx = DDX[temp_id,ccd_id];
    variable dy = DDY[temp_id,ccd_id];
    
    % let's find the min/max chip numbers
    variable xmin, xmax, ymin, ymax;
    ( xmin, xmax ) = minmax( x );
    ( ymin, ymax ) = minmax( y );

    % which gives the min/max "grid" numbers
    variable nx_lo, nx_hi, ny_lo, ny_hi;
    nx_lo = int(xmin/dx);
    nx_hi = int(xmax/dx) + 1;
    ny_lo = int(ymin/dy);     
    ny_hi = int(ymax/dy) + 1;

    % draw the grid lines: do vertical then
    % horizontal lines    
    % NOTE: 
    %   I am NOT bothered about off-by-one errors
    %   when drawing these lines. However, for the 
    %   counts it should be correct
    %
    variable i, ctr, x1, y1, x2, y2;
    y1 = ny_lo*dy; 
    y2 = ny_hi*dy;

    % x axis
    x1 = nx_lo * dx;
    ctr = 1;
    for( i = nx_lo; i <= nx_hi; x1 += dx, i++, ctr++ ) {
	() = chips_line( x1, y1, x1, y2 );
	() = chips_eval( "ln " + string(ctr) + " RED" );
    }

    x1 = nx_lo*dx; 
    x2 = nx_hi*dx;

    % y axis
    y1 = ny_lo * dy;
    for( i = ny_lo; i <= ny_hi; y1 += dy, i++, ctr++ ) {
	() = chips_line( x1, y1, x2, y1 );
	() = chips_eval( "ln " + string(ctr) + " RED" );
    }

    % display the lines
    chips_redraw;

    % now count the number of photons in each region - we 
    % store in a linked list so that we don't have to write 
    % a sorting routine
    % (nb chipx, chipy are integer types)
    %
    variable j, index_y, xx, index_x, num, ntotal;

    y1 = ny_lo * dy + 1;
    y2 = y1 + dy - 1;

    ntotal = 1.0 * length(x);  % ensure a real number

    % points to start of the linked list
    variable head = NULL;

    for ( j = ny_lo; j < ny_hi; y1+=dy, y2+=dy, j++ ) {

        % extract only those events with y values in the required range
        index_y = where( y >= y1 & y <= y2 );
        xx = x[index_y];

	x1 = nx_lo * dx + 1;
	x2 = x1 + dx - 1;

        for ( i = nx_lo; i < nx_hi; x1+=dx, x2+=dx, i++ ) {

            % extract only those events with x values in the required range
            index_x = where( xx >= x1 & xx <= x2 );
	    num = length( index_x );

	    % store the results IF there are any photons in the region
            if ( num ) {
                % create a structure to store the data
		variable region = struct { photons, fraction, region, next, prev };
		region.photons  = num;
		region.fraction = sprintf( "%6.4f", num / ntotal );
		region.region   = sprintf( "%04d:%04d %04d:%04d", x1, x2, y1, y2 );

		head = slist_insert( head, region );

	    } % if num

        } % i: nx_lo .. (nx_hi-1)

    } % j: ny_lo .. (ny_hi-1)

    % print out the results to the screen
    output( "#         Region           Num    Fraction" );
    output( "#    chipx     chipy     photons" );
    variable elem;
    foreach ( head ) {
	elem = ();
	voutput( "   %s  %8d  %s", elem.region, elem.photons, elem.fraction );
    }

} % draw_grid()

%
% driver routine: this is what the user calls
%
define display_fef_regions() { % filename (incl region and chip_id filter if necessary)

    if ( _NARGS != 1 )
	usage( "display_fef_regions(filename)\nwhere filename includes any region and ccd_id filters required." );

    % pop variables off the stack
    variable filename = ();

    % avoid reading in columns we do not need
    voutput( "# Reading points from %s", filename );
    variable points = readfile( filename + "[cols time,ccd_id,chipx,chipy]" );
    if ( points == NULL )
        error( "** Stopped as unable to read event list." );

    % check we only have one ccd_id value
    variable ccd_min, ccd_max;
    ( ccd_min, ccd_max ) = minmax( points.ccd_id );
    if ( ccd_min != ccd_max )
	error( "** Error: region covers more than one chip.\n          Please include a filter on ccd_id in the filename." );

    voutput( "# Chip is ACIS-%d", ccd_min );

    % find the temperature
    variable tstart = points.time[0];
    variable temp, temp_id;
    ( temp, temp_id ) = find_temperature( tstart );
    if ( temp == NULL ) return;
    voutput( "# Focal Plane temperature = %d C", temp );
    voutput( "# Number of photons in region = %d", points._nrows );

    % plot the points
    % -- we autoscale to the data, rather than display the whole
    %    chip
    variable old_redraw = chips_auto_redraw(0);
    variable old_chips  = @chips;

    chips_clear;
    chips.symbolstyle = _chips->point;

    () = curve( points.chipx, points.chipy );
%%    () = chips_set_xrange( -10, 1035 );
%%    () = chips_set_yrange( -10, 1035 );
    () = chips_eval( "xlabel 'CHIP X'" );
    () = chips_eval( "ylabel 'CHIP Y'" );

    % protect any underscores in the filename
    variable title;
    ( title, ) = strreplace( filename, "_", "\\_", strlen(filename) );
    () = chips_eval( "title 'ACIS-" + string(ccd_min) + " " + title + "'" );
    () = chips_eval( "title size 1.2" );

    % now draw the lines and calculate the numbers in each region
    draw_grid( ccd_min, temp_id, points.chipx, points.chipy );

    () = chips_auto_redraw(old_redraw);
    set_state( "chips", old_chips );

} % display_fef_regions()

