%
% $Id: setplot.sl,v 1.3 2004/11/02 16:19:23 dburke Exp $
%
% Requires:
%   CIAO 3.2
%
% Aim:
%   Provides an interactive mechanism for setting plot attributes
%   - ie Sherpa will prompt you for the attributes in a similar
%     manner to CIAO tools prompt for parameters
%
% Usage:
%  Ensure that setplot.par is available (ie 'paccess setplot'
%  returns a file name) and then
%
%    sherpa> require("setplot")
%
%  or - in your ~/.sherparc
%
%    require("setplot");
%
%  and then 
%
%   sherpa> setplot("data")
%
%  or  
%
%   sherpa> setplot("data","xlog+")
%
% Functions defined:
%   setplot( plotname [, option1, option2, ... ] )
%     Sets the plot options for the given plot name
%
% Requires:
%   setplot.par in PFILES path (i.e. can plist/pset/punlearn setplot)
%
% Aim:
%   Provides an alternative interface to the Sherpa
%   plot state object (eg see 'ahelp sherpa.plot', ...) for
%   setting plot options
%
%   Here we use a parameter interface to set parameter values
%   - mainly so that we can let the parameter interface deal
%     with querying the values rather than dealing with I/O
%     ourselves
%
%   The aim is to have a function that is easy to use for the
%   casual user: people using these functions many times will
%   likely find that this interface gets in the way, since it
%   always asks for parameters. Such users are suggested to
%   use a different interface.
%
% Notes:
%   This is similar to paramest.sl - and is in fact based on v1.9
%   of that code - EXCEPT for one *major* difference:
%     we take the default values from the relevant sherpa.* object
%     rather than the parameter file
%   This is because we want to use the same parameter names for
%   different plots (eg plot, dataplot, resplot all share the same
%   fields) but the default values can be different which is hard
%   to do sensibly in the parameter interface (without having
%   parameters for each structure which is pointless). Also - the
%   idea of setting parameter values using the parameter tools
%   looks like it is less efficient than its worth so I do not think
%   it is a big upheaval.
%
%   If the user breaks out of the routine (control-c) then there is 
%   a possibility that strange things might happen with the parameter
%   interface when the routine is called again. We attempt to work
%   around the latter issue by having a "global" variable (at least 
%   global to this file/compilation unit) which contains the
%   Param_File_Type value for the last opened file. Then when a
%   function is called we check that this parameter is NULL and - 
%   if not - close it. This seems to work.
%
%   The use of the parameter library to access information is
%   not a necessity here; we use it since it provides an interface
%   that is familiar to users AND means we do not have to write
%   our own input/output routines for querying the user (although
%   these would not be too hard).
%
% To Do:
%   - on a 'usage error' list the parameters used by a function
%     so the user can find them out just by calling the function
%     with no parameters
%

%
% This is based on v1.9 of paramest.sl but the parameter handling is
% a lot simpler since we do not change the min/max/prompt fields
% here.
%

require("sherpa");
require("paramio");

% gives the name of the parameter file
% - check we can find it
% (could do this only once but easiest to do it every time,
%  and now we use require/provide it is even less of a worry)
%  
private variable PARNAME = "setplot";
if ( NULL == paccess(PARNAME) )
  verror( "\nError: unable to find the parameter file for %s\n",
          PARNAME );

private variable fp = NULL;

% utility routine to close the parameter file if it was open
%
private define _close_fp() {
  if ( NULL != fp ) {
    paramclose ( fp );
    fp = NULL;
  }
}

% can we access the parameter file?
%
private define _open_paramfile (args) {
  if ( NULL == args ) fp = paramopen( PARNAME, "rwL" );
  else                fp = paramopen( PARNAME, "rwL", args );
  if ( NULL == fp )
    verror( "\nERROR: Unable to open the parameter file '%s'\n", PARNAME );
} % _open_paramfile

% handle the optional parameters
% - if optional arguments are given then we need to
%   create an argument array for paramopen() that
%   begins with the "tool" name
%
private define _process_args () {
  !if ( _NARGS ) return NULL;

  variable args = String_Type [_NARGS+1];
  args[0] = PARNAME;

  variable i = 1;
  foreach ( __pop_args(_NARGS) ) {
    variable arg = ();
    args[i] = arg.value;
    i++;
  }

  return args;
} % _process_args()

%
% Currently a no-op: intended to allow the structure being edited to
% be restored to the state it was in prior to the function being called
% [should be controlled by a parameter value]
%
% low priority to implement
%
%%private define _restore_state() {}


%
% the following functions do the actual work: there is one
% for each "plot type" which asks the user for the parameters
% of interest to it and sets the necessary fields
%

% sherpa.plot/sherpa.data/sherpa.resplot structure:
%   x_errorbars      =  0
%   y_errorbars      =  0
%   errs_style       =  bar
%   errs_type        =  both
%   x_log            =  0
%   y_log            =  0
%   curvestyle       =  step
%   curvecolor       =  default
%   symbolstyle      =  none
%   symbolcolor      =  default
%   symbolsize       =  2
%   xlabel_size      =  1.5
%   ylabel_size      =  1.5
%   zlabel_size      =  1.5
%   title_size       =  1.5
%   tickvals_size    =  1.5
%   prefunc          =  NULL
%   postfunc         =  NULL
%
% We do NOT ask for/set the pre/postfunc hooks
%
private define _setplot_plot_state(args,s) {

  if ( sherpa_Plot_State != typeof(s) )
    verror( "Internal error: %s() called with object of type %S",
             _function_name, typeof(s) );

  % set up the default values given the fields
  % in the supplied structure
  % - use a mode of rwLH (OTT?) to stop automatic parameters
  %   being prompted for
  %
  _close_fp();
  fp = paramopen( PARNAME, "rwLH" );
  if ( NULL == fp ) verror( "\nERROR: Unable to open param file %s\n", PARNAME );
  pputb(   fp, "xlog",         s.x_log );
  pputb(   fp, "ylog",         s.y_log );
  pputb(   fp, "xerror",       s.x_errorbars );
  pputb(   fp, "yerror",       s.y_errorbars );
  pputstr( fp, "errorstyle",   s.errs_style );
  pputstr( fp, "errortype",    s.errs_type );
  pputstr( fp, "curvestyle",   s.curvestyle );
  pputstr( fp, "curvecolor",   s.curvecolor );
  pputstr( fp, "symbolstyle",  s.symbolstyle );
  pputstr( fp, "symbolcolor",  s.symbolcolor );
  pputf(   fp, "symbolsize",   s.symbolsize );
  pputf(   fp, "xlabelsize",   s.xlabel_size );
  pputf(   fp, "ylabelsize",   s.ylabel_size );
  pputf(   fp, "zlabelsize",   s.zlabel_size );
  pputf(   fp, "titlesize",    s.title_size );
  pputf(   fp, "tickvalssize", s.tickvals_size );
  paramclose(fp);
  fp = NULL;

  % can we access the parameter file?
  _open_paramfile( args );

  % we let the parameter library cast from boolean to 0/1

  %% "Important" axis options
  %
  s.x_log = pgetb( fp, "xlog" );
  s.y_log = pgetb( fp, "ylog" );

  %% Errors
  %
  s.x_errorbars = pgetb( fp, "xerror" );
  s.y_errorbars = pgetb( fp, "yerror" );
  if ( s.x_errorbars or s.y_errorbars ) {
    s.errs_style = pgetstr( fp, "errorstyle" );
    s.errs_type  = pgetstr( fp, "errortype" );
  }

  %% Line connecting points
  %
  s.curvestyle = pgetstr( fp, "curvestyle" );
  if ( 0 != strcmp( "noline", s.curvestyle ) )
    s.curvecolor = pgetstr( fp, "curvecolor" );

  %% Symbols at points
  %
  s.symbolstyle = pgetstr( fp, "symbolstyle" );
  if ( 0 != strcmp( "none", s.symbolstyle ) ) {
    s.symbolcolor = pgetstr( fp, "symbolcolor" );
    s.symbolsize  = pgetf( fp, "symbolsize" );
  }

  %% "Less important" axis options
  %
  % - do we need zlabel size for all plot types?
  %
  s.xlabel_size = pgetf( fp, "xlabelsize" );
  s.ylabel_size = pgetf( fp, "ylabelsize" );
  s.zlabel_size = pgetf( fp, "zlabelsize" );

  s.title_size    = pgetf( fp, "titlesize" );
  s.tickvals_size = pgetf( fp, "tickvalssize" );

  % close the parameter file
  _close_fp;

} % _setplot_plot_state()

private define _setplot_plot(a)     { _setplot_plot_state( a, sherpa.plot ); }
private define _setplot_dataplot(a) { _setplot_plot_state( a, sherpa.dataplot ); }
private define _setplot_resplot(a)  { _setplot_plot_state( a,sherpa.resplot ); }

% sherpa.fitplot structure:
%   x_errorbars      =  0
%   y_errorbars      =  1
%   errs_style       =  bar
%   errs_type        =  both
%   x_log            =  1
%   y_log            =  1
%   curvestyle       =  noline
%   curvecolor       =  default
%   symbolstyle      =  square
%   symbolcolor      =  default
%   symbolsize       =  2
%   fit_curvestyle   =  step
%   fit_curvecolor   =  red
%   fit_symbolstyle  =  none
%   fit_symbolcolor  =  default
%   fit_symbolsize   =  2
%   xlabel_size      =  1.5
%   ylabel_size      =  1.5
%   zlabel_size      =  1.5
%   title_size       =  1.5
%   tickvals_size    =  1.5
%   prefunc          =  NULL
%   postfunc         =  NULL
%
% (sherpa.plot/etc + the fit_XXX fields)
%
% We do NOT ask for/set the pre/postfunc hooks
%
private define _setplot_fitplot(args) {

  variable s = sherpa.fitplot;

  % set up the default values given the fields
  % in the supplied structure
  % - use a mode of rwLH (OTT?) to stop automatic parameters
  %   being prompted for
  %
  _close_fp();
  fp = paramopen( PARNAME, "rwLH" );
  if ( NULL == fp ) verror( "\nERROR: Unable to open param file %s\n", PARNAME );
  pputb(   fp, "xlog",         s.x_log );
  pputb(   fp, "ylog",         s.y_log );
  pputb(   fp, "xerror",       s.x_errorbars );
  pputb(   fp, "yerror",       s.y_errorbars );
  pputstr( fp, "errorstyle",   s.errs_style );
  pputstr( fp, "errortype",    s.errs_type );
  pputstr( fp, "curvestyle",   s.curvestyle );
  pputstr( fp, "curvecolor",   s.curvecolor );
  pputstr( fp, "symbolstyle",  s.symbolstyle );
  pputstr( fp, "symbolcolor",  s.symbolcolor );
  pputf(   fp, "symbolsize",   s.symbolsize );

  pputstr( fp, "fitcurvestyle",   s.fit_curvestyle );
  pputstr( fp, "fitcurvecolor",   s.fit_curvecolor );
  pputstr( fp, "fitsymbolstyle",  s.fit_symbolstyle );
  pputstr( fp, "fitsymbolcolor",  s.fit_symbolcolor );
  pputf(   fp, "fitsymbolsize",   s.fit_symbolsize );

  pputf(   fp, "xlabelsize",   s.xlabel_size );
  pputf(   fp, "ylabelsize",   s.ylabel_size );
  pputf(   fp, "zlabelsize",   s.zlabel_size );
  pputf(   fp, "titlesize",    s.title_size );
  pputf(   fp, "tickvalssize", s.tickvals_size );
  paramclose(fp);
  fp = NULL;

  % can we access the parameter file?
  _open_paramfile( args );

  % we let the parameter library cast from boolean to 0/1

  %% "Important" axis options
  %
  s.x_log = pgetb( fp, "xlog" );
  s.y_log = pgetb( fp, "ylog" );

  %% Errors
  %
  s.x_errorbars = pgetb( fp, "xerror" );
  s.y_errorbars = pgetb( fp, "yerror" );
  if ( s.x_errorbars or s.y_errorbars ) {
    s.errs_style = pgetstr( fp, "errorstyle" );
    s.errs_type  = pgetstr( fp, "errortype" );
  }

  %% Line connecting points
  %
  s.curvestyle = pgetstr( fp, "curvestyle" );
  if ( 0 != strcmp( "noline", s.curvestyle ) )
    s.curvecolor = pgetstr( fp, "curvecolor" );

  %% Symbols at points
  %
  s.symbolstyle = pgetstr( fp, "symbolstyle" );
  if ( 0 != strcmp( "none", s.symbolstyle ) ) {
    s.symbolcolor = pgetstr( fp, "symbolcolor" );
    s.symbolsize  = pgetf( fp, "symbolsize" );
  }

  %% Line/symbol for the model
  %
  s.fit_curvestyle = pgetstr( fp, "fitcurvestyle" );
  if ( 0 != strcmp( "noline", s.fit_curvestyle ) )
    s.fit_curvecolor = pgetstr( fp, "fitcurvecolor" );
  s.fit_symbolstyle = pgetstr( fp, "fitsymbolstyle" );
  if ( 0 != strcmp( "none", s.fit_symbolstyle ) ) {
    s.fit_symbolcolor = pgetstr( fp, "fitsymbolcolor" );
    s.fit_symbolsize  = pgetf( fp, "fitsymbolsize" );
  }

  pputstr( fp, "fitcurvestyle",   s.fit_curvestyle );
  pputstr( fp, "fitcurvecolor",   s.fit_curvecolor );
  pputstr( fp, "fitsymbolstyle",  s.fit_symbolstyle );
  pputstr( fp, "fitsymbolcolor",  s.fit_symbolcolor );
  pputf(   fp, "fitsymbolsize",   s.fit_symbolsize );

  %% "Less important" axis options
  %
  % - do we need zlabel size for all plot types?
  %
  s.xlabel_size = pgetf( fp, "xlabelsize" );
  s.ylabel_size = pgetf( fp, "ylabelsize" );
  s.zlabel_size = pgetf( fp, "zlabelsize" );

  s.title_size    = pgetf( fp, "titlesize" );
  s.tickvals_size = pgetf( fp, "tickvalssize" );

  % close the parameter file
  _close_fp;

} % _setplot_fitplot()

%
% this variable provides the mapping between the plot name and the
% function to call to set up its parameters
%
private variable _allowed_plots = Assoc_Type [Ref_Type];

_allowed_plots["data"]    = &_setplot_dataplot();
_allowed_plots["back"]    = &_setplot_dataplot();
_allowed_plots["udata"]   = &_setplot_dataplot();
_allowed_plots["uback"]   = &_setplot_dataplot();
_allowed_plots["dcounts"] = &_setplot_dataplot();
_allowed_plots["bcounts"] = &_setplot_dataplot();

_allowed_plots["delchi"]     = &_setplot_resplot();
_allowed_plots["ratio"]      = &_setplot_resplot();
_allowed_plots["residuals"]  = &_setplot_resplot();
_allowed_plots["bdelchi"]    = &_setplot_resplot();
_allowed_plots["bratio"]     = &_setplot_resplot();
_allowed_plots["bresiduals"] = &_setplot_resplot();

_allowed_plots["fit"]     = &_setplot_fitplot();
_allowed_plots["ufit"]    = &_setplot_fitplot();
_allowed_plots["bfit"]    = &_setplot_fitplot();
_allowed_plots["bufit"]   = &_setplot_fitplot();
_allowed_plots["backfit"] = &_setplot_fitplot();

_allowed_plots["errors"]      = &_setplot_plot();
_allowed_plots["berrors"]     = &_setplot_plot();
_allowed_plots["syserrors"]   = &_setplot_plot();
_allowed_plots["bsyserrors"]  = &_setplot_plot();
_allowed_plots["staterrors"]  = &_setplot_plot();
_allowed_plots["bstaterrors"] = &_setplot_plot();
_allowed_plots["model"]       = &_setplot_plot();
_allowed_plots["mcounts"]     = &_setplot_plot();
_allowed_plots["bmodel"]      = &_setplot_plot();
_allowed_plots["bmcounts"]    = &_setplot_plot();
_allowed_plots["chi squ"]     = &_setplot_plot();
_allowed_plots["bchisqu"]     = &_setplot_plot();
_allowed_plots["statistic"]   = &_setplot_plot();
_allowed_plots["bstatistic"]  = &_setplot_plot();
_allowed_plots["weight"]      = &_setplot_plot();
_allowed_plots["bweight"]     = &_setplot_plot();
_allowed_plots["filter"]      = &_setplot_plot();
_allowed_plots["bfilter"]     = &_setplot_plot();
_allowed_plots["group"]       = &_setplot_plot();
_allowed_plots["bgroup"]      = &_setplot_plot();

_allowed_plots["arf"]         = &_setplot_plot();
_allowed_plots["barf"]        = &_setplot_plot();
_allowed_plots["expmap"]      = &_setplot_plot();
_allowed_plots["bexpmap"]     = &_setplot_plot();

_allowed_plots["source"]      = &_setplot_plot();
_allowed_plots["background"]  = &_setplot_plot();
_allowed_plots["bg"]          = &_setplot_plot();
_allowed_plots["noise"]       = &_setplot_plot();
_allowed_plots["bnoise"]      = &_setplot_plot();
_allowed_plots["kernel"]      = &_setplot_plot();

_allowed_plots["psf"]         = &_setplot_plot();
_allowed_plots["bpsf"]        = &_setplot_plot();

% should make this a prettier list
private variable _allowed_plot_names =
  strjoin( assoc_get_keys(_allowed_plots), " " );

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Usage:
%   setplot( plotname )
%   setplot( plotname, val1, ..., valN )
%
% Aim:
%   Sets the fields appropriate to the given plot, where
%   plotname is the string you would use in 'lplot data' or
%   'lplot delchi' or ...
%
%   plotname can be one of:
%     data or dcounts
%     back or bcounts
%   ...
% 
%   The remaining parameters are read from the 'setplot'
%   parameter file and depend on the value of plotname.
%
%   If more than one argument is supplied the extra are
%   assumed to specify parameter values in the form
%   "name=value" - ie as a string. The name must be supplied
%
%   The default values for the prompts are taken from the
%   *state object* and NOT the parameter file. This breaks
%   the similarity with tools, but seems much-more useful
%   in this context. It is also different to v1.9 of
%   paramest.sl
%
% Examples:
%   Sets up the options for 'lplot data' (and similar plots)
%
%     sherpa> setplot("data")
%
%   Here the xlog parameter is set to yes and it will NOT be
%   prompted for.
%
%     sherpa> setplot("data","xlog+")
%
define setplot() {

  if ( _NARGS < 1 )
    usage(
	"setplot( plotname [, arg1, ..., argN ] )\n\n" +
	"where plotname is one of:\n  " +
        _allowed_plot_names + "\n"
         );

  % process the command-line
  variable alist = __pop_args(_NARGS-1);
  variable args = _process_args( __push_args(alist) );

  % check that the plot is known. It would be nice to
  % use "assoc_key_exists(_allowed_plots,plotname)" but
  % we want to allow users to use substrings and it
  % still work (since you can say 'lplot resid').
  % The following will not work if we have a valid plot name that
  % is a substring of another plot name.
  %
  variable plotname = ();
  variable regexp   = "^" + strlow( plotname ); % do comparison in lower case

  variable func = NULL;
  foreach ( assoc_get_keys(_allowed_plots) ) {

    % does this name match the user-supplied name
    variable name = ();

    if ( 1 == string_match( name, regexp, 1 ) ) {
      if ( NULL != func )
        verror( "Error: '%s' does not have a unique match in:\n  %s\n",
                plotname, _allowed_plot_names );
      func = _allowed_plots[name];
    }

  } % foreach: assoc_get_keys

  % did we get a match
  if ( NULL == func )
    verror( "Error: no match found for '%s' in:\n  %s\n",
            plotname, _allowed_plot_names );
    
  % get the parameters appropriate for the plot type
  %
  @func( args );

  vmessage( "\nThe parameters for %s (and related) plots has been set.", plotname );
  vmessage( "Try 'lplot %s'\n", plotname );

} % setplot()

provide( "setplot" );

%%%
%%% End
%%%
