%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% $Id: sherpa_plotfns.sl,v 1.29 2004/11/02 15:59:14 dburke Exp $
%
% Thread:
%   http://cxc.harvard.edu/ciao/threads/sherpa_slangplot/
%   http://chandra.ledas.ac.uk/ciao/threads/sherpa_slangplot/
%
% Aim:
%   See how the new plotting capabilities in Sherpa (introduced in
%   CIAO 3.0) help us when customising Sherpa plots.
%
% Requires:
%   CIAO 3.2
%
% Functions:
%   plot_rprof() - new for CIAO 3.0.2
%   plot_rprofr()
%   plot_rprofd()
%     When you have a fit to an image, this function plots the
%     radial profile of the data and best fit, with a radial
%     profile of the residual image underneath (so somewhat
%     like 'lplot fit resid' for a radial profile fit)
%   - plot_rprof() and plot_rprofr() are the same (the former is
%     kept for backwards compatability) and plot the residuals in counts
%     (e.g. 'lplot resid')
%   - plot_rprofd() plots the residuals in counts/errors - ie sigma
%     (e.g. 'lplot delchi')
%
%   plot_eprof() - new for CIAO 3.0.2
%   plot_eprofr()
%   plot_eprofd()
%     Similar to plot_rprof() but uses elliptical annuli
%     rather than circular ones.
%
%   lpfr([n])
%     a short form of 'lplot 2 fit [n] resid [n]'
%
%   lpfd([n])
%     a short form of 'lplot 2 fit [n] delchi [n]'
%
% Variables:
%   This package defines 2 variables that are 'exported' into the global
%   namespace:
%     _sherpa_plotfns_version_string
%     _sherpa_plotfns_version
%   they define the version of the file, and are analogous to the
%   _slang_version and _slang_version_string variables, except
%   that they are not read only, and the numbering scheme is
%   a.b not a.b.c
%
% Other changes:
%   The sherpa plot configuration variables (aka state objects)
%   are changed to make 'lplot 2 fit resid/delchi' plots look
%   better (at least in my opinion). this involves adding
%   pre/postfunc hooks in the sherpa.multiplot config variable.
%
% Notes:
%  . Since the prefunc can not find out everything about the
%    requested plot it can not "avoid" situations like
%      sherpa> lplot 2 fit 1 resid 2
%    or - a perhaps more common plot -
%      sherpa> lplot 3 fit 1 resid 2 arf 3
%
%    In CIAO 3.2 we could use the sherpa.multiplot.pre/postfunc
%    hooks to only do the processing when necessary. It would
%    probably work like
%      sherpa.multiplot.prefunc works out whether this is
%      an 'interesting' plot specification and, if so,
%      sets up the resplot.pre/postfunc hooks.
%      sherpa.multiplot.postfunc would then remove the
%      resplot hooks.
%
%    This has not been implemented since no users have complained
%    about the current functionality.
%  
% History:
%   2002/04/25 - added lpfd()
%   2002/06/05 - take advantage of sherpa_utils functions if they exist
%   2003/07/24 - Major re-work for CIAO 3.0: do not really need lpfr/lpfd
%                anymore but left in for backwards compatability
%   2003/08/11 - fix lpfd() to plot delchi and not resid
%   2003/11/13 - added plot_rprof() and dbg_sherpa_plotfns() [undocumented]
%   2003/11/18 - added _sherpa_plotfns_version[_string] variables
%   2003/12/31 - We now *require* sherpa_utils.sl
%   2004/01/06 - plot_rprof draws a line indicating 'y=0' line in residual plot
%   2004/01/29 - added ability to specify model name or x/y values to
%                plot_rprof
%   2004/02/02 - to celebrate the second super bowl for the Patriots,
%                - the radial profiles now include error estimates
%                - plot_eprof() has been added; this calculates the profile
%                  in elliptical annuli
%   2004/03/18 - to celebrate almost a week since the Chandra deadline,
%                added the temporary functions (need a better name/scheme)
%                  plot_rprofd() 
%                  plot_eprofd() 
%                which plot the residuals in units of "sigma"
%                NOTE: the residual plot is NO LONGER normalised by area
%   2004/05/28 - minor clean-up of names:
%                  plot_rprofr() plots residuals in counts ('lplot resid')
%                  plot_rprofd() plots residuals in sigma (counts/error) ('lplot delchi')
%                and plot_rprof() is a "short-form" for plot_rprofr(), kept
%                around for backwards compatibility. ditto for the "eprof" versions
%   2004/08/10 - fixed a bug in _sherpa_resid_prefunc() which showed up when
%                a model was fit to the  background (rather than it being subtracted)
%   2004/11/02 - Updated for CIAO 3.2; added a provide("sherpa_plotfns"); line
%                and use require for loading CIAO modules. Use get_fit() rather
%                than get_goodness().
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

require("sherpa");

% version checks
%
if ( 30200 > sherpa_version )
  verror( stderr, "ERROR:\n sherpa_plotfns.sl requires at least\n version 3.2 of Sherpa\n" );

if ( 0 == is_defined("_sherpa_utils_version") ) {
  require("sherpa_utils");
  if ( 0 == is_defined("_sherpa_utils_version") )
    verror( "ERROR:\n Unable to load sherpa_utils.sl from sherpa_plotfns.sl\n" );
}

% Usage:
%  ( strval, intval ) = _get_version_info( rcs_version_string )
%
% Aim:
%   Parse the RCS version string into string and integer values
%
%   rcs_version_string should be "@Revision@" [where @ is replaced
%   by $ and the whole string has been expanded by RCS].
%
private define _get_version_info() {
  if ( 1 != _NARGS ) {
    variable cleanup = __pop_args( _NARGS );
    usage( "(strval,intval) = _get_version_info( rcs_version_string )\n" +
      " where the input string looks like '$Revision: 1.29 $'.\n" );
    return ( NULL, NULL );
  }

  variable rcs = ();

  % strip out the excess text
  variable version_string = substr( rcs, 12, strlen(rcs)-13 );

  % calculate a numeric value
  variable version_int = 0;
  variable i = 0;
  foreach ( strchopr( version_string, '.', 0 ) ) {
    variable num = ();
    version_int += ( integer(num) * 10^i );
    i+=2;
  }

  % return
  return ( version_string, int(version_int) );

} % _get_version_info()

% set up the version variables from the RCS Revision id
% - unlike the slang version variables (_slang_version and _slang_version_string)
%   these variables are writeable
public variable _sherpa_plotfns_version_string, _sherpa_plotfns_version;
( _sherpa_plotfns_version_string, _sherpa_plotfns_version ) =
  _get_version_info( "$Revision: 1.29 $" );

% the debug state can be toggled using set_dbg_sherpa_plotfns()
%  - send it 1 for on and 0 for off, no argument to do nothing
%  - returns the old setting
%  - this is for debugging work and is not really useful for
%    the user
%
% The "__is_initialized" check ensures that the value is not
% reset when the script is re-loaded (only useful when testing
% since the general user will use require to load, which avoids the
% need for this check)
%
private variable _dbg;
!if ( __is_initialized(&_dbg) ) _dbg = 0;
public define set_dbg_sherpa_plotfns() {
  switch ( _NARGS )
  { case 0: return _dbg; }
  { case 1:
      % NOTE: we do not check that the input variable is sensible
      variable nval = ();
      variable oval = _dbg;
      _dbg = nval;
      return oval;
  }
  {
    () = __pop_args(_NARGS); 
    usage( "dbg_sherpa_plotfns( [0 or 1] );" );
  }
} % set_dbg_sherpa_plotfns()

private define dbg_sherpa_plotfns(msg) {
  if ( _dbg ) () = fprintf( stderr, "DBG [sherpa_plotfns]: %s\n", msg );
}

%
% Usage:
%   ce( command );
%   ce( [ command1, command2, ... ] );
%
% Aim:
%   Send ChIPS a command - or sequence of commands if the argument
%   is a string array - to be executed. We *IGNORE* the return value
%   of each command (ie we do not care if it succeeds or not).
%
private define ce( commands ) {
  if ( Array_Type != typeof(commands) ) commands = [ commands ];
  foreach ( commands ) {
    variable comm = ();
    dbg_sherpa_plotfns( "Sending ChIPS '" + comm + "'" );
    () = chips_eval( comm );
  }
} % ce()

%
% Takes in the dataset number and the centre of the profile (xcen,ycen),
% ellipticity (ellip) and angle (theta, in radians)
% and returns a 1D array giving the square of the distance
% of the pixel centre from the profile centre.
%
% the coordinate systems of axes (as returned by
% get_axes) and (xcen,ycen) are assumed to be the same
%
% We only care about the centre of each pixel -- ie
% we only include whole pixels into a radial bin,
% and use the centre of the pixel to decide where it goes
%
private define axes2rdist(dnum,xcen,ycen,ellip,theta) {

  variable axes = get_axes(dnum);
  if ( NULL == axes ) {
    () = fprintf( stderr, "Error: unable to get axis information for dataset %d\n", dnum );
    return NULL;
  }

  % get the mid-points of the axes
  %
  variable xmid, ymid;
  if ( NULL == axes[0].mid ) {

    xmid = 0.5 * ( axes[0].lo + axes[0].hi );
    ymid = 0.5 * ( axes[1].lo + axes[1].hi );

  } else {
    xmid = axes[0].mid;
    ymid = axes[1].mid;
  }

  % return the "distance^2" array
  % (follow the labels used in 'ahelp beta2d')
  %
  variable dx = xmid - xcen;
  variable dy = ymid - ycen;

  variable x_new = dx * cos(theta) + dy * sin(theta);
  variable y_new = dy * cos(theta) - dx * sin(theta);

  variable efact = (1.0 - ellip) * (1.0 - ellip);

  return ( x_new*x_new*efact + y_new*y_new ) / efact;

} % axes2rdist()

% Usage:
%   modelname = guess_pos_names( dnum, 0 );
%   modelname = guess_pos_names( dnum, 1 );
%
% If the second argument is 0 then look for components
% which contain parameters called xpos and ypos.
% If the second argument is 1 then the components must
% also contain ellip and theta parameters.
%
% dnum is the source dataset to check. The return value is
% the name of the compoenent (e.g. foo from "beta2d[foo]")
%
% if no model is found then calls verror() - ie does not return
% anything
%
% takes advantage of get_source_components() from sherpa_utils.sl
%
private define guess_pos_names( dnum, eflag ) {

  % want a list of models used in the source expression 
  %
  variable models = get_source_components( dnum );
  dbg_sherpa_plotfns( "Source " + string(dnum) +
		      " has the following source components\n " +
		      strjoin(models,", ") );

  % loop through and look for xpos/ypos parameters - we could do it based on the type
  % of the model - ie hard-code in the info, but that
  % will not handle user-defined models. Plus this way is as easy/not too time consuming
  %
  % if there are any duplicated models we re-process them (it is not
  % worth checking to avoid this)
  %
  variable names = Assoc_Type [];
  names[ "xpos" ] = 1;
  names[ "ypos" ] = 1;
  if ( eflag ) {
    names[ "ellip" ] = 1;
    names[ "theta" ] = 1;
  }

  variable matches = Assoc_Type [Int_Type];
  variable nmatch = 2 + eflag*2;
  foreach ( models ) {
    variable name = ();
    variable pars = get_par( name );
    if ( NULL == pars )
      verror( "Error: unable to find parameters for model '%s'\n", name );

    variable flag = 0;
    foreach ( pars ) {
      variable par = ();
      variable parname = strtok( par.name, "." )[1];
      if ( assoc_key_exists(names,parname) ) flag++;
    } % foreach: pars

    if ( nmatch == flag ) matches[name] = 1;
  } % foreach: models

  variable n = length( matches );

  if ( n == 1 ) {
    name = assoc_get_keys(matches)[0];
    if ( eflag )
      dbg_sherpa_plotfns( "Found xpos/ypos/ellip/theta parameters for model " + name );
    else
      dbg_sherpa_plotfns( "Found xpos/ypos parameters for model " + name );
    return name;
  }

  variable emsg;
  if ( 0 == n ) {
    emsg = sprintf( "Unable to find any model in '%s'\nwhich has ",
                     get_source_expr(dnum) );
    if ( eflag )
      emsg += "xpos, ypos, ellip, and theta parameters\n";
    else
      emsg += "xpos and ypos parameters\n";

  } else {

    if ( eflag )
      emsg = "The following models contain xpos, ypos, ellip, and theta parameters:\n";
      else
        emsg = "The following models contain xpos and ypos parameters:\n";

    emsg += "  " + strjoin( assoc_get_keys(matches), ", " ) +
            "\nPlease specify which component to use.\n";
	    
  }

  emsg +=
    "\nSee the ahelp file for the command (or enter the command with no\n" +
    "arguments) to see how to specify the component name to use\n\n";
  error( emsg );

} % guess_pos_names()

% ( xlo, xhi, xscale ) = get_xaxis_info(pane);
%
% find the x-axis range for the given pane
% - changes the current pane to #pane
%
% Note:
%  in CIAO 3.0/3.1/3.2 the x-axis range will be returned as the
%  logarithm of the values if the scale is logarithmic
%  (really should return the linear values since that is how we
%   set the axis), so we need to re-scale them if necessary
%
private define get_xaxis_info(pane) {
  () = chips_set_pane(pane);
  variable xs = chips_get_xscale;
  variable xlo, xhi;
  ( xlo, xhi ) = chips_get_xrange();
  if ( 0 == xs ) {
    dbg_sherpa_plotfns( "Resetting x-axis range from log to linear values [pane " + string(pane) + "]" );
    xlo = 10^xlo;
    xhi = 10^xhi;
  }
  return ( xlo, xhi, xs );

} % get_xaxis_info()

% create the radial-profile plots using ChIPS
%
% - much of the plot configuration is taken from the sherpa.fitplot and
%   sherpa.resplot structures except:
%     a) the y axis of the resid plot is left in a linear scale
%     b) the pre/postfunc hooks are ignored
%
% if flag is 1 then the residuals are plotted "as is",
% otherwise we plot residuals/error
%
private define do_rplot( dnum, xcen, ycen, hist, flag ) {

  % find the source expression
  %
  variable sexpr = get_source_expr( dnum );
  if ( orelse { NULL == sexpr } { "" == sexpr } )
    verror( "Error: unable to find the source expression for dataset %d\n", dnum );
  dbg_sherpa_plotfns( "Source expression: " + sexpr );

  % use %.2f for centre since should be "okay" for logical/physical
  % coordinates - and it is only meant to give a hint
  % - this is not really a sensible long-term solution
  % - and does not really scale well for elliptical plots
  %
  variable title = sprintf( "%s @ %.2f,%.2f", sexpr, xcen, ycen );

  % plot the data
  %
  variable oflag = chips_auto_redraw(0);
  chips_clear();

  % set up the draing area; use the same pane sizes as for the
  % residual-plot override above
  %
  () = chips_split(1,2);
  ce( [ "d 1 location 0.15 0.9 0.4 0.9",
        "d 2 location 0.15 0.9 0.1 0.4",
        "title '" + title + "'",
        "title size " + string(sherpa.fitplot.title_size) ] );

  %% Plot 1: data + model
  %
  () = chips_set_pane(1);

  % only draw error bars along y axis
  %
  () = curve( hist.x, hist.y, hist.dy );
  ce( [ "symbol " + sherpa.fitplot.symbolstyle,
        "symbol " + sherpa.fitplot.symbolcolor,
        "symbol size " + string(sherpa.fitplot.symbolsize),
        "errs " + sherpa.fitplot.errs_style,
        "errs " + sherpa.fitplot.errs_type,
        sherpa.fitplot.curvestyle,
        sherpa.fitplot.curvecolor ] );

  () = curve( hist.x, hist.model );
  ce( [ "symbol " + sherpa.fitplot.fit_symbolstyle,
        "symbol " + sherpa.fitplot.fit_symbolcolor,
        "symbol size " + string(sherpa.fitplot.fit_symbolsize),
        sherpa.fitplot.fit_curvestyle,
        sherpa.fitplot.fit_curvecolor ] );

  % axes
  if ( sherpa.fitplot.x_log ) ce( "log x" );
  if ( sherpa.fitplot.y_log ) ce( "log y" );

  ce( [ "ylabel '\\Sigma counts/area'",
        "ylabel size " + string(sherpa.fitplot.ylabel_size),
        "tickvals size " + string(sherpa.fitplot.tickvals_size),
        "tickvals x off" ] );

  % get the limits (needed to work around an issue with
  % drawing y=0 line in residual plot in CIAO 3.0.2/3.1/3.2)
  %
  variable xlo, xhi;
  ( xlo, xhi, ) = get_xaxis_info(1);

  %% Plot 2: residuals
  %
  % if flag=1 we plot residuals
  % otherwise         residuals/sigma
  %
  % NOTE: have to re-normalise since we do NOT want to plot
  %   "per area" values. This is a new change in version 1.26
  %
  () = chips_set_pane(2);

  % draw on the 'y=0' line and do it before drawing anything else
  % (so that everything else draws on top of this line)
  %
  % To make it draw across the whole display we need to set the x limits
  % rather than allow the limits to be calcualted automatically. This
  % is a bug/feature in ChIPS 3.0.2/3.1/3.2
  %
  () = chips_set_xrange( xlo, xhi );
  () = chips_line( xlo, 0.0, xhi, 0.0 );

  if ( flag ) {
    () = curve( hist.x, hist.resid * hist.npix, hist.dy * hist.npix );
  } else {
    % as both hist.resid and hist.dy are "per pixel" we
    % remove the per-area scaling in the division
    %
    () = curve( hist.x, hist.resid / hist.dy, 1.0+hist.dy*0.0 );
  }

  ce( [ "symbol " + sherpa.resplot.symbolstyle,
        "symbol " + sherpa.resplot.symbolcolor,
        "symbol size " + string(sherpa.resplot.symbolsize),
        "errs bar",
        "errs " + sherpa.resplot.errs_style,
        "errs " + sherpa.resplot.errs_type,
        sherpa.resplot.curvestyle,
        sherpa.resplot.curvecolor ] );

  ce( [ "xlabel 'radius (" + get_coord(dnum) + " pixels)'",
        "title '" + title + "'",
        "xlabel size " + string(sherpa.resplot.xlabel_size),
        "ylabel size " + string(sherpa.resplot.ylabel_size),
        "tickvals size " + string(sherpa.resplot.tickvals_size) ] );

  if ( flag )
    ce( "ylabel '\\Sigma residuals'" );
  else
    ce( "ylabel '(\\Sigma residuals)/\\sigma'" );

  % axes
  if ( sherpa.resplot.x_log ) ce( "log x" );
  %%%if ( sherpa.fitplot.y_log ) ce( "log y" );

  () = chips_auto_redraw(oflag);
  chips_redraw();

} % do_rplot()

%
% - need to think what to do for error analysis if data has had a background
%   dataset subtracted from it. currently we print a warning message and
%   ignore the errors from the background subtraction
%
% - we use the Gehrels upper-limit approximation to calculate the errors
%   could do both upper and lower limits (ie unequal), or use the approximations
%   from the paper by Harald Ebeling
%
private define create_radial_histogram(dnum,rmin,rmax,rwidth,xc,yc,ellip,theta) {

  % sort out the distance array
  %
  variable rdist2 = axes2rdist( dnum, xc, yc, ellip, theta );
  if ( NULL == rdist2 ) return;
  dbg_sherpa_plotfns( "Created distance scale" );

  % get the image array
  %
  variable image = get_data(dnum);
  if ( NULL == image )
    verror( "Error: unable to get data for dataset %d\n", dnum );
  dbg_sherpa_plotfns( "got the data" );

  % get the model array
  %
  variable model = get_mcounts(dnum);
  if ( NULL == model )
    verror( "Error: unable to get the model for dataset %d\n", dnum );
  dbg_sherpa_plotfns( "got the model" );

  % get the residual array
  %
  variable resid = get_residuals(dnum);
  if ( NULL == resid )
    verror( "Error: unable to get residual data for dataset %d\n", dnum );
  dbg_sherpa_plotfns( "got the residual data" );

  % warning message as currently undecided what to do in this
  % case - I expect it to be infrequent for Chandra & XMM data
  %
  if ( is_subtracted(dnum) )
    () = fprintf( stdout,
      "WARNING:\n\n" +
      "Errors are calculated using the Gehrels approximataion and under\n" +
      "the assumption that the data has not been background subtracted.\n" +
      "This dataset (%d) has been background subtracted so the errors will\n" +
      "be underestimated.\n\n", dnum );

  variable nbins = int( (rmax - rmin) / rwidth );
  variable rlo = rmin + [0:nbins-1]*rwidth;
  variable rhi = rlo + rwidth;
  dbg_sherpa_plotfns(sprintf("Original num of bins in residual profile = %d", nbins) );

  variable rlo2 = rlo * rlo;
  variable rhi2 = rhi * rhi;
  dbg_sherpa_plotfns( sprintf( "array types: rdist2 %S rlo2 %S rhi2 %S",
                _typeof(rdist2),_typeof(rlo2),_typeof(rhi2)) ); 

  variable hist_data  = Double_Type [nbins];
  variable hist_err   = Double_Type [nbins];
  variable hist_model = Double_Type [nbins];
  variable hist_resid = Double_Type [nbins];
  variable hist_npix  = Int_Type [nbins];
  variable flag       = Int_Type [nbins];

  variable i, index;
  foreach ( [0:nbins-1] ) {
    i = ();
    dbg_sherpa_plotfns( sprintf("Profile bin: %d  rlo2= %f  rhi2= %f", i, rlo2[i],rhi2[i]) );

    index = where( rdist2 >= rlo2[i] and rdist2 < rhi2[i] );
    variable nindex = length(index);
    dbg_sherpa_plotfns( "  #bins found = " + string(nindex) );

    % we use the Gehrels approximation for errors (upper-limit)
    % - I would hope that these bins have a large number of counts
    %   in, so we could just use the Gaussian approximation
    %   but I do not want to assume this
    %
    if ( 0 != nindex ) {
      variable npts = sum( image[index] );
      hist_data[i]  = npts / nindex;
      hist_err[i]   = ( 1.0 + sqrt(npts + 0.75) ) / nindex;
      hist_model[i] = sum( model[index] ) / nindex;
      hist_resid[i] = sum( resid[index] ) / nindex;
      hist_npix[i]  = nindex;
      flag[i] = 1;
    }
  }

  % strip out those bins with no data
  %
  index = where( 1 == flag );
  rlo = rlo[index];
  rhi = rhi[index];
  hist_data  = hist_data[index];
  hist_err   = hist_err[index];
  hist_model = hist_model[index];
  hist_resid = hist_resid[index];
  hist_npix  = hist_npix[index];

  variable rmid = 0.5*(rlo+rhi);

  variable s = struct { x, y, dy, model, resid, npix };
  set_struct_fields( s, rmid, hist_data, hist_err, hist_model,
		     hist_resid, hist_npix );
  return s;

} % create_radial_histogram()

% Usage:
%   (xc,yc) = find_rprof_centre( usage, args, dset );
%
% Aim:
%   given a set of parameters from the user, find the
%   centre we should be using for the radial profile.
%
%   usage is an array of strings which give the usage info
%   for the main routine
%   args is an array of structs, each contianing a value field
%   with the user-supplied argument in it.
%
%   We expect one of:
%   - args is empty; find centre ourselves
%   - args contains 1 element - the name of the model
%   - args contains 2 elements
%     - if strings, the names of the x/y parameters
%     - otherwise, the x/y values to use
%
private define find_rprof_centre( _usage_error, args, dset ) {

  variable xname, yname;

  dbg_sherpa_plotfns(
    sprintf("find_rprof_centre() sent %d args to process",length(args))
	    );
  switch( length(args) )
  {
  case 0:
    variable mname = guess_pos_names( dset, 0 );
    xname = mname + ".xpos";
    yname = mname + ".ypos";
  }
  {
  case 1:
    if ( String_Type != typeof(args[0].value) )
      @_usage_error;
    dbg_sherpa_plotfns( sprintf("User asked for position from model = %s",
				args[0].value) );
    xname = args[0].value + ".xpos";
    yname = args[0].value + ".ypos";
  }
  {
  case 2:
    variable t1 = typeof( args[0].value );
    variable t2 = typeof( args[1].value );
    if ( t1 == String_Type and t1 != t2 )
      @_usage_error();
    if ( t1 == String_Type ) {
      % user has given parameter names
      %
      xname = args[0].value + ".xpos";
      yname = args[1].value + ".ypos";
    } else {
      % user has given the actual position, which is easy to process
      % [ensure they are Double_Type by multiplying by 1.0]
      %
      return ( 1.0 * args[0].value, 1.0 * args[1].value );
    }
  }
  { @_usage_error(); }

  % if we have got this far then xname/yname should contain the
  % names of the parameters containing the x/y positions we want
  %
  variable xpar = get_par(xname);
  if ( NULL == xpar )
    verror( "Error: %s does not appear to be a valid parameter name\n", xpar );
  variable ypar = get_par(yname);
  if ( NULL == ypar )
    verror( "Error: %s does not appear to be a valid parameter name\n", ypar );

  return ( xpar.value, ypar.value );

} % find_rprof_centre()

%
% Usage:
%   sherpa.resplot.prefunc  = &_sherpa_resid_prefunc;
%   sherpa.resplot.postfunc = &_sherpa_resid_postfunc;
%
% Aim:
%   Apply customisations to plots that match
%     "lplot 2 fit resid"
%   or - more generally -
%     "lplot 2 plottype n resid/delchi/ratio n"
%
%   In these plots we FORCE the residual plot to have a
%   linear y scale, no matter what the setting of
%   sherpa.resplot.y_log. This is because "set_log" also
%   changes the sherpa.resplot.y_log field and we do not
%   want to make the user have to reset it each time
%   This is done in the postfunc hook since we need the
%   plot to have been drawn at this point.
%
%   We also ensure that the first two panes have a common x range
%   (the minimum rather than maximum so that "lp 2 arf resid"
%    does not have a lot of blank area in the residual plot)
%   in the postfunc hook
%
% Note:
%   We try to make the hook smart enough not to do anything
%   when it should not - eg during
%     "lplot resid"
%   However, we can not make it foolproof, as noted in the
%   "Notes" section at the head of the file
%

static define _sherpa_resid_prefunc(data, label, dataset, type) {

  dbg_sherpa_plotfns( "Called: _sherpa_resid_prefunc() for plot #" + string(chips_get_pane()) );

  % assume we only want the re-sizing happening IF
  % the residuals are being drawn in the second pane
  %  - this does NOT correctly handle situations like
  %    "lplot 3 fit resid arf"
  %
  if ( 2 != chips_get_pane() ) return;

  % set up the title
  % - we add the reduced chi-square value IF it is defined
  %
  variable model  = get_source_expr(dataset);
  variable title  = sprintf( "title '%s", model );

  % If fitres is an array then we assume that we have
  % the values for the combined, source, and background datasets
  % and that we want to display the value of the separate datasets
  %
  variable fitres = get_fit(dataset);
  if ( Array_Type == typeof(fitres) ) {
    if ( NULL != fitres[0].rstat ) {
      if ( 3 == length(fitres) ) {
	title += sprintf( " %s \\chi^2_r= %f  %s \\chi^2_r= %f",
			  fitres[1].datatype, fitres[1].rstat,
			  fitres[2].datatype, fitres[2].rstat );
      } else {
	% just in case something strange has happened
	%
	title += sprintf( " %s \\chi^2_r= %f", fitres[0].datatype, fitres[0].rstat );
      }
    }
  } else if ( NULL != fitres.rstat ) {
    title += sprintf( " \\chi^2_r= %f", fitres.rstat );
  }
  title += "'";

  % make the second plot smaller (and remove the gap between the plots)
  %
  ce( [ title,
        "d 1 location 0.15 0.9 0.4 0.9",
        "d 2 location 0.15 0.9 0.1 0.4" ] );
  
  % remove the x axis values from the first plot
  () = chips_set_pane(1);
  ce( "tickvals x off" );

  % as we are a pre-func and the previous command has moved us 
  % back to the first pane we need to go back to the second
  % pane (or else the residuals would be drawn in the first pane)
  %
  () = chips_set_pane(2);

} % _sherpa_resid_prefunc

% cleans up the setting of the y_log field back to its user-selected
% value AND ensures that the first two panels have the same
% axis range. Note that if one is in logarithmic spacing
% and the other linear we leave as is.
%
static define _sherpa_resid_postfunc() {

  dbg_sherpa_plotfns( "Called: _sherpa_resid_postfunc() for plot #" + string(chips_get_pane()) );

  if ( 2 != chips_get_pane() ) return;

  % ignore the sherpa.resplot.y_log setting
  () = chips_set_yscale(1);

  % ensure that the x-axis ranges of the first two panes
  % are the same (even if they do not both have the same scale)
  % In CIAO 3.0/3.1/3.2 chips_get_xscale() does not "ignore" the scale
  % setting for the axis so we have to convert the values for
  % logged axes
  %
  variable xs1, xlo1, xhi1;
  ( xlo1, xhi1, xs1 ) = get_xaxis_info(1);

  variable xs2, xlo2, xhi2;
  ( xlo2, xhi2, xs2 ) = get_xaxis_info(2);

  % could find out if we need to do this but easier just to do it
  % - originally used the largest range but that can result in
  %   plots with a lot of space, so restrict to the smaller range
  %   (which will likely be the residual part)
  %
  variable xlo = max([xlo1,xlo2]);
  variable xhi = min([xhi1,xhi2]);
  () = chips_set_pane(1);
  () = chips_set_xrange(xlo,xhi);
  () = chips_set_pane(2);
  () = chips_set_xrange(xlo,xhi);

  % Ensure the two panes have the same X-axis scale - we take it
  % from the main plot in case there is a difference.
  % It is not a perfect solution - the scaling is not as good as
  % when the x_log fields match, but it just makes sure the plots
  % are usable
  %
  () = chips_set_xscale(xs1);

} % _sherpa_resid_postfunc

%%%%%%%%%%%%%%%%%%%%
%%
%% apply the customisations
%%
%% it is just a matter of setting fields in the relevant
%% configuration variable (aka state object)
%%

%% set up the hooks for residual plots
sherpa.resplot.prefunc  = &_sherpa_resid_prefunc;
sherpa.resplot.postfunc = &_sherpa_resid_postfunc;

% customise the residual plots
sherpa.resplot.x_errorbars = 0;
sherpa.resplot.symbolstyle = "bigpoint";
sherpa.resplot.symbolsize  = 1;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% public routines
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%
% Usage:
%   lpfr
%   lpfr()
%   lprf(n)
%
% Aim:
%   Plots fit and residual
%
% This is no longer needed (it was, prior to CIAO 3.0), but is left in
% for backwards compatability
%
public define lpfr () {

  % are we called correctly?
  variable n = 1;
  switch ( _NARGS )
  { case 0: }
  { case 1: n = (); }
  { () = __pop_args(_NARGS); usage( _function_name + "( [dataset number] )\n" ); }

  % if it fails the default error message is good enough
  () = sherpa_eval( "lplot 2 fit " + string(n) + " resid " + string(n) );

} % lpfr()

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%
% Usage:
%   lpfd
%   lpfd()
%   lpfd(n)
%
% Aim:
%   Plots fit and delchi
%
% This is no longer needed (it was, prior to CIAO 3.0), but is left in
% for backwards compatability
%
public define lpfd () {

  % are we called correctly?
  variable n = 1;
  switch ( _NARGS )
  { case 0: }
  { case 1: n = (); }
  { () = __pop_args(_NARGS); usage( _function_name + "( [dataset number] )\n" ); }

  % if it fails the default error message is good enough
  () = sherpa_eval( "lplot 2 fit " + string(n) + " delchi " + string(n) );

} % lpfd()

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%
% Usage:
%   plot_rprofr( dnum, xname, yname, min, max, step )
%   plot_rprofr( dnum, xval,  yval,  min, max, step )
%   plot_rprofr( dnum, modelname,    min, max, step )
%   plot_rprofr( dnum,               min, max, step )
%
%   plot_rprofr( xname, yname, min, max, step )
%   plot_rprofr( xval,  yval,  min, max, step )
%   plot_rprofr( modelname,    min, max, step )
%   plot_rprofr(               min, max, step )
%
%   where
%     dnum is the dataset number - defaults to 1
%     xname is the name of the X position parameter
%     yname is the name of the Y position parameter
%     xval  is the X coordinate of the center of the radial profile
%     yval  is the Y coordinate of the center of the radial profile
%
%     min is the minimum edge of the radial bins
%     max is the maximum edge of the radial bins
%     step is the width of each bin
%     - the coordinates are defined by the COORD setting for
%       the dataset
%
% Usage:
%   plot_eprofr( dnum, modelname, min, max, step )
%   plot_eprofr( dnum,            min, max, step )
%
%   plot_eprofr( modelname, min, max, step )
%   plot_eprofr(            min, max, step )
%
%   where
%     dnum is the dataset number - defaults to 1
%     modelname is the name of the model containing xpos,ypos,ellip,theta
%       parameters
%
%     min is the minimum edge of the radial bins
%     max is the maximum edge of the radial bins
%     step is the width of each bin
%     - the coordinates are defined by the COORD setting for
%       the dataset
%
% The plot_rprof() and plot_eprof() commands are identical to
% plot_rprofr() and plot_eprofr(), for backwards compatability.
%
% The plot_rprofd() and plot_eprofd() commands plot the same data
% as the plot_rprofr() and plot_eprofr() commands, except that the
% residuals are in units of sigma (counts / error)
%
% Requires:
%   >= CIAO 3.0.2
%
% Aim:
%   Try to plot up radial profiles of the
%     data and model
%     residuals
%   from a 2D fit within Sherpa. The look of the plot can be controlled
%   by changing fields in the sherpa.fitplot and sherpa.resplot
%   structures (although not everything is used, e.g. we always make the
%   residual plots have a linear y axis and the error bar settings are
%   ignored). 
%
%   Only those radial bins with pixels in them will be plotted.
%
%   plot_rprofr() uses circular bins whereas plot_eprofr() uses the
%   best-fit ellipticity and angle of the model to generate ellipses
%
% Example (assumes sherpa_plotfns.sl is available in the search path):
%
%   sherpa> require("sherpa_plotfns.sl")
%   sherpa> data img.fits
%   sherpa> paramprompt off
%   sherpa> coord physical
%   sherpa> source = beta2d[clus]
%   sherpa> fit
%   sherpa> image fit
%   sherpa> thaw clus.ellip clus.theta
%   sherpa> clus.theta = 2
%   sherpa> clus.ellip = 0.5
%   sherpa> fit
%   sherpa> image fit
%   sherpa> plot_eprofr( 0, 80, 5 )
%
% Enhancement ideas:
%   - one possibility is to let the user specify a set of regions
%     on the image/residual map (using ds9) which we then use
%     to calculate the profile - ie instead of assuming regularly-spaced
%     circular annuli
%     which leads to being able to set up such regions from the
%     sherpa command line (at least so that they pick up the correct
%     x and y centre values).
%

private define plot_rprof_usage(funcname) {

  % @@ will be replaced by funcname below
  %
  variable usage_str =
    [
     "There are a number of ways that @@() can be called:",
     "  @@( [dnum,] min, max, step );",
     "  @@( [dnum,] modelname, min, max, step );",
     "  @@( [dnum,] xval, yval, min, max, step );",
     "  @@( [dnum,] xname, yname, min, max, step );",
     "where",
     "  dnum is the dataset number - defaults to 1", 
     "  min is the minimum edge of the radial bins",
     "  max is the maximum edge of the radial bins",
     "  step is the width of each bin",
     "",
     "The center of the profile is - if not specified - taken from the source",
     "component which contains both xpos and ypos parameters. If the source",
     "model contains more than one such component (e.g. if you are using two",
     "beta models) then you must specify which center to use. This can be done",
     "by one of:",
     " - giving the name of the component (the modelname parameter above)",
     " - giving the names of the model parameters to use for the X and Y",
     "   centers; e.g. @@(\"core.xpos\",\"core.ypos\",0,200,5);'",
     " - giving the X and Y coordinates; e.g. @@(4230.45,3685.8,0,200,5);",
     "",
     "All coordinates are expected to be in the coordinate system defined",
     "by the COORD command (or set_coord() function) for the dataset.",
     ""
     ];

  % it appears that there is a limit to the length of the string
  % that error() can display, so we do the text separately
  % and then give error a small string.
  %
  % have hard-coded the number of replacements to 10; this is currently
  % larger than the number of @@ pairs in the string
  %
  ( usage_str, ) = strreplace( strjoin(usage_str,"\n"), "@@", funcname, 10 );
  () = printf( "%s\n", usage_str );
  error( funcname + "() called incorrectly." );

} % plot_rprof_usage

private define _plot_rprof( funcname, resid_type, args ) {

  dbg_sherpa_plotfns( sprintf("number of arguments = %d",length(args)) );

  % we use the type of the first argument to decide whether the user has
  % specified the dataset or not. Unfortunately it is not as simple as
  % I had hoped since I want to allow plot_rprof(0,200,5)
  % and 0 is an integer here - so we remove the rmin/max/delta first
  % and then check for everything else
  %
  if ( length(args) < 3 or length(args) > 6 )
    plot_rprof_usage(funcname);

  % process the rmin/max/width first
  %
  variable rmin   = args[-3].value;
  variable rmax   = args[-2].value;
  variable rwidth = args[-1].value;
  args = args[[0:-4]];
  dbg_sherpa_plotfns( sprintf("Radial bins = %f : %f : %f",rmin,rmax,rwidth) );

  % if the first argument is NOT a string or float/double then assume
  % it is the dataset number (slightly cleaner than trying to check for
  % all integer types). 
  %
  % the valid possibilities are:
  %   1  dnum
  %   2  dnum, modelname [string]
  %   3  dnum, xname [string], yname [string]
  %   3  dnum, xpos [number], ypos [number]
  % and
  %   0  no arguments
  %   1  modelname [string]
  %   2  xname [string], yname [string]
  %   2  xpos [number], ypos [number]
  %
  variable dnum;
  switch ( length(args) )
  { case 0: dnum = 1; }
  {
  case 1:
    if ( String_Type == typeof(args[0].value) )
      dnum = 1;
    else {
      dnum = args[0].value;
      args = args[[1:-1]];
    }
  }
  {
  case 2:
    variable t1 = typeof(args[0].value);
    variable t2 = typeof(args[1].value);
    % 
    % looking for one of:
    %   dnum, modelname
    %   xname, yname
    %   xpos, ypos
    % so we use the logic that if the types are not the same
    % AND the second value is a string then we must have
    % "dnum, modelname". If we did not include the check on
    % the second arg being a string then we would have problems
    % if the user supplies xpos as an integer and ypos as a float
    % (or vice-versa)
    %
    if ( t1 != t2 and t2 == String_Type ) {
      dnum = args[0].value;
      args = args[[1:-1]];
    } else
      dnum = 1;
  }
  {
  case 3: 
    dnum = args[0].value;
    args = args[[1:-1]];
  }
  { verror( "Internal error in %s - missed argument case", _function_name ); }
  dbg_sherpa_plotfns( sprintf("dataset number = %d",dnum) );

  variable xcen, ycen;
  ( xcen, ycen ) = find_rprof_centre( &plot_rprof_usage, args, dnum );
  dbg_sherpa_plotfns( sprintf("X/Y position = %f %f",xcen,ycen) );

  % now we can create the histograms
  %
  variable hist = create_radial_histogram( dnum, rmin, rmax, rwidth,
					   xcen, ycen, 0.0, 0.0 );

  % plot the data
  %   resid_type = 1 => residuals in units of counts
  %   resid_type = 0 => residuals in units of counts/error (ie sigma)
  %
  do_rplot( dnum, xcen, ycen, hist, resid_type );

} % _plot_rprof()

% kept in for backwards compatibility, it is the same as plot_rprofr()
%
public define plot_rprof() {
  dbg_sherpa_plotfns( sprintf("in %s with %d arguments",_function_name(),_NARGS) );
  variable args = __pop_args( _NARGS );
  _plot_rprof( _function_name(), 1, args );
} % plot_rprof()

public define plot_rprofr() {
  dbg_sherpa_plotfns( sprintf("in %s with %d arguments",_function_name(),_NARGS) );
  variable args = __pop_args( _NARGS );
  _plot_rprof( _function_name(), 1, args );
} % plot_rprofr()

public define plot_rprofd() {
  dbg_sherpa_plotfns( sprintf("in %s with %d arguments",_function_name(),_NARGS) );
  variable args = __pop_args( _NARGS );
  _plot_rprof( _function_name(), 0, args );
} % plot_rprofr()

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

private define plot_eprof_usage(funcname) {

  % @@ will be replaced by funcname below
  %
  variable usage_str =
    [
     "There are a number of ways that @@() can be called:",
     "  @@( [dnum,] min, max, step );",
     "  @@( [dnum,] modelname, min, max, step );",
     "where",
     "  dnum is the dataset number - defaults to 1", 
     "  modelname is the model to use for the profile shape", 
     "  min is the minimum edge of the radial bins",
     "  max is the maximum edge of the radial bins",
     "  step is the width of each bin",
     "",
     "The profile is defined by the xpos, ypos, ellip, and theta parameters",
     "of the source model. If the source model contains only one component",
     "with xpos and ypos parameters then you do not need to specify the name",
     "of this component. If the source contains more than one component - e.g.",
     "if you are using two beta2d models - then you must specify which component",
     "is to be used via the modelname argument.",
     "",
     "All coordinates are expected to be in the coordinate system defined",
     "by the COORD command (or set_coord() function) for the dataset.",
     ""
     ];

  % it appears that there is a limit to the length of the string
  % that error() can display, so we do the text separately
  % and then give error a small string
  %
  % have hard-coded the number of replacements to 10; this is currently
  % larger than the number of @@ pairs in the string
  %
  ( usage_str, ) = strreplace( strjoin(usage_str,"\n"), "@@", funcname, 10 );
  () = printf( "%s\n", usage_str );
  error( funcname + "() called incorrectly." );

} % plot_eprof_usage

private define _plot_eprof( funcname, resid_type, args ) {

  dbg_sherpa_plotfns( sprintf("number of arguments = %d",length(args)) );

  %
  % #Num  arguments   
  %    5  dnum, modelname, min, max, step
  %    4  dnum, min, max, step
  %    4  modelname, min, max, step
  %    3  min, max, step
  %
  if ( length(args) < 3 or length(args) > 5 )
    plot_eprof_usage(funcname);

  % process the rmin/max/width first
  %
  variable rmin   = args[-3].value;
  variable rmax   = args[-2].value;
  variable rwidth = args[-1].value;
  args = args[[0:-4]];
  dbg_sherpa_plotfns( sprintf("Radial bins = %f : %f : %f",rmin,rmax,rwidth) );

  variable dnum, modelname;
  switch ( length(args) )
  {
  case 0:
    dnum = 1;
    modelname = guess_pos_names( dnum, 1 );
  }
  {
  case 1:
    if ( String_Type == typeof(args[0].value) ) {
      dnum = 1;
      modelname = args[0].value;
    } else {
      dnum = args[0].value;
      modelname = guess_pos_names( dnum, 1 );
    }
  }
  {
  case 2:
    % could also check that arg1 is an integer (but then also have
    % to check for ULong/Long/UInt/...)
    %
    if ( String_Type != typeof(args[1].value) )
      plot_eprof_usage();

    dnum      = args[0].value;
    modelname = args[1].value;
  }
  { verror( "Internal error in %s - missed argument case", _function_name ); }
  dbg_sherpa_plotfns( sprintf("dataset number = %d",dnum) );
  dbg_sherpa_plotfns( sprintf("modelname = %s",modelname) );

  variable pars = Assoc_Type [Double_Type];
  foreach ( [ "xpos", "ypos", "ellip", "theta" ] ) {
    variable pname = ();
    variable par = get_par( modelname + "." + pname );
    if ( NULL == par )
      verror( "Error: %s does not appear to be a valid parameter name for model %s\n", pname, modelname );
    pars[pname] = par.value;
  }

  variable xcen  = pars["xpos"];
  variable ycen  = pars["ypos"];
  variable ellip = pars["ellip"];
  variable theta = pars["theta"];

  dbg_sherpa_plotfns( sprintf("X/Y position = %f %f",xcen,ycen) );
  dbg_sherpa_plotfns( sprintf("ellip = %f  theta = %f",ellip,theta) );

  % now we can create the histograms
  %
  variable hist = create_radial_histogram( dnum, rmin, rmax, rwidth,
					   xcen, ycen, ellip, theta );

  % plot the data
  %   resid_type = 1 => residuals in units of counts
  %   resid_type = 0 => residuals in units of counts/error (ie sigma)
  %
  do_rplot( dnum, xcen, ycen, hist, resid_type );

} % _plot_eprof()

% kept in for backwards compatibility, it is the same as plot_eprofr()
%
public define plot_eprof() {
  dbg_sherpa_plotfns( sprintf("in %s with %d arguments",_function_name(),_NARGS) );
  variable args = __pop_args( _NARGS );
  _plot_eprof( _function_name(), 1, args );
} % plot_eprof()

public define plot_eprofr() {
  dbg_sherpa_plotfns( sprintf("in %s with %d arguments",_function_name(),_NARGS) );
  variable args = __pop_args( _NARGS );
  _plot_eprof( _function_name(), 1, args );
} % plot_eprofr()

public define plot_eprofd() {
  dbg_sherpa_plotfns( sprintf("in %s with %d arguments",_function_name(),_NARGS) );
  variable args = __pop_args( _NARGS );
  _plot_eprof( _function_name(), 0, args );
} % plot_eprofr()

provide ("sherpa_plotfns");

%%%
%%% End
%%%
