%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% sherpa_simspec.sl
%
% Version 0.1 (2004/01/28)
%
% Author:  Christopher J. Stawarz (cstawarz@cfa.harvard.edu)
%
% Purpose:
%   Provides fakeit_norm(), which generates a simulated dataset with a
%   user-specified normalization.  Also provides a number of convenience
%   functions that are meant to simplify the scripting of simulations.
%
% Public Constants:
%   sherpa_simspec->NORM_NONE
%   sherpa_simspec->NORM_EFLUX
%   sherpa_simspec->NORM_PFLUX
%   sherpa_simspec->NORM_COUNTS
%   sherpa_simspec->NORM_COUNT_RATE
%
%   sherpa_simspec->PARAMEST_NONE
%   sherpa_simspec->PARAMEST_UNC
%   sherpa_simspec->PARAMEST_COV
%   sherpa_simspec->PARAMEST_PROJ
%
% Public Functions:
%   write_data_pha()
%   write_mdl()
%   set_batchon()
%   set_batchoff()
%   write_lplot()
%
%   set_fakeit_time()
%   set_fakeit_backscale()
%   run_fakeit()
%
%   sherpa_simspec->normtype_string_to_value()
%   fakeit_norm()
%
%   sherpa_simspec->paramest_method_string_to_value()
%   fit_paramest()
%
% Message and Error Functions:
%   By default, the package will use message() for debugging/informational
%   messages (verbosity level is controlled by sherpa_simspec->verbose;
%   value of 0 means no messages, 5 means maximum messages) and verror()
%   for error messages.
%
%   However, it is possible to customize message/error handling by assigning
%   function references to the variables sherpa_simspec->message_func and
%   sherpa_simspec->error_func, respectively.  If these variables are set,
%   then the arguments to the private functions do_message() and do_error()
%   will be passed to the specified handlers, and the default functions will
%   not be called.
%
%   If used, the function associated with sherpa_simspec->message_func must
%   accept one or two arguments (a message string, or an integer verbosity
%   value and message string).  The function associated with
%   sherpa_simspec->error_func must accept one argument (an error message
%   string).
%
% History:
%   2004/01/28 - initial version
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


implements("sherpa_simspec");

%_traceback=1;
%_debug_info=1;


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Message and error functions
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


% Message verbosity level
static variable verbose = 0;

% Package name to report in error messages
private variable package_name = current_namespace();

% References to external message and error functions
static variable message_func = NULL;
static variable error_func = NULL;


private define do_message()
%
% Display message string "msg".  If optional integer "vlevel" is provided,
% display message only if verbose >= vlevel.
%
% Alternatively, if "message_func" is not NULL, pass all arguments to
% the function it refers to and return.  (That is, use a user-specified
% external message handler.)
%
{
    if (message_func != NULL) {
	variable args = __pop_args(_NARGS);
	@message_func(__push_args(args));
	return;
    }

    variable vlevel = 0;
    variable msg;

    switch (_NARGS)
      { case 1 : msg = (); }
      { case 2 : (vlevel,msg) = (); }

    if (verbose >= vlevel)  message(msg);
}


private define do_error()
%
% Issue error message "msg" in standard format and throw error. 
%
% Alternatively, if "error_func" is not NULL, pass all arguments to
% the function it refers to and return.  (That is, use a user-specified
% external error handler.)
%
{
    if (error_func != NULL) {
	variable args = __pop_args(_NARGS);
	@error_func(__push_args(args));
	return;
    }

    variable msg = ();
    verror("# %s: ERROR: %s",package_name,msg);
}


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Public convenience functions
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


public define write_data_pha()
%
% Wrapper for "WRITE DATA [#] <filename> PHA"
%
{
    variable dataset = NULL;
    variable outfile;

    switch (_NARGS)
      { case 1 : outfile = (); }
      { case 2 : (dataset,outfile) = (); }
      { usage(sprintf("status = %s([dataset,]filename)",_function_name())); }

    if (dataset == NULL)
        dataset = 1;
    else if (typeof(dataset) != Integer_Type)
	do_error("dataset number must be an integer");

    outfile = string(outfile);

    if (-1 == sherpa_eval(sprintf("WRITE DATA %d %s PHA",dataset,outfile)))
        return 0;

    return 1;
}


public define write_mdl()
%
% Wrapper for "WRITE MDL <filename>"
%
{
    variable outfile;

    switch (_NARGS)
      { case 1 : outfile = (); }
      { usage(sprintf("status = %s(filename)",_function_name())); }

    outfile = string(outfile);

    % Work around bug that causes 'write mdl' to create two files
    % ("MDL_Data" and "MDL_Models") if specified file exists
    if (NULL != stat_file(outfile))
	if (0 != system("rm -f " + outfile))  return 0;

    if (-1 == sherpa_eval(sprintf("WRITE MDL \"%s\"",outfile))) return 0;
    return 1;
}


public define set_batchon()
%
% Wrapper for "BATCH ON"
%
{
    () = chips_eval("BATCH ON");
}


public define set_batchoff()
%
% Wrapper for "BATCH OFF"
%
{
    () = chips_eval("BATCH OFF");
}


public define write_lplot()
%
% Wrapper for
%
%   LPLOT <plotargs>
%   PRINT POSTFILE <outfile>
%
{
    variable plotargs;
    variable outfile;

    switch (_NARGS)
      { case 2 : (outfile,plotargs) = (); }
      { usage(sprintf("status = %s(filename,plotargs)",_function_name())); }

    outfile = string(outfile);
    plotargs = string(plotargs);

    if (-1 == sherpa_eval(sprintf("LPLOT %s",plotargs)))  return 0;
    if (-1 == chips_eval(sprintf("PRINT POSTFILE %s",outfile)))  return 0;

    return 1;
}


public define set_fakeit_time()
%
% Wrapper for "FAKEIT [#] TIME = <time>"
%
{
    variable dataset = NULL;	% Dataset number
    variable exptime;		% Exposure time to set

    switch (_NARGS)
      { case 1 : exptime = (); }
      { case 2 : (dataset,exptime) = (); }
      { usage(sprintf("status = %s([dataset,]time)",_function_name())); }

    if (dataset == NULL)
        dataset = 1;
    else if (typeof(dataset) != Integer_Type)
	do_error("dataset number must be an integer");

    exptime = typecast(exptime,Double_Type);

    if (-1 == sherpa_eval(sprintf("FAKEIT %d TIME = %g",dataset,exptime)))
        return 0;

    return 1;
}


public define set_fakeit_backscale()
%
% Wrapper for "FAKEIT [#] BACKSCALE = <backscale>"
%
{
    variable dataset = NULL;	% Dataset number
    variable backscale;		% Background scale factor to set

    switch (_NARGS)
      { case 1 : backscale = (); }
      { case 2 : (dataset,backscale) = (); }
      { usage(sprintf("status = %s([dataset,]backscale)",_function_name())); }

    if (dataset == NULL)
        dataset = 1;
    else if (typeof(dataset) != Integer_Type)
	do_error("dataset number must be an integer");

    backscale = typecast(backscale,Double_Type);

    if (-1 == sherpa_eval(sprintf("FAKEIT %d BACKSCALE = %g",
      dataset,backscale)))
        return 0;

    return 1;
}


public define run_fakeit()
%
% Wrapper for "FAKEIT [#]"
%
{
    variable dataset = NULL;	% Dataset number

    switch (_NARGS)
      { case 0 : ; }
      { case 1 : dataset = (); }
      { usage(sprintf("status = %s([dataset])",_function_name())); }

    if (dataset == NULL)
        dataset = 1;
    else if (typeof(dataset) != Integer_Type)
	do_error("dataset number must be an integer");

    if (-1 == sherpa_eval(sprintf("FAKEIT %d",dataset)))  return 0;
    return 1;
}


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Simulation-related functions
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


% Normalization types (defining them as functions makes them act as constants,
% i.e. users can use but not change their values)
static define NORM_NONE() { return 0; }
static define NORM_EFLUX() { return 1; }
static define NORM_PFLUX() { return 2; }
static define NORM_COUNTS() { return 3; }
static define NORM_COUNT_RATE() { return 4; }

% Minimum and maximum normalization-type values
private define NORM_TYPE_MIN() { return 0; }
private define NORM_TYPE_MAX() { return 4; }


private define is_valid_normtype(normtype)
%
% Returns true if integer "normtype" corresponds to a valid normalization
% type.
%
{
    return ((normtype >= NORM_TYPE_MIN) and (normtype <= NORM_TYPE_MAX));
}


static define normtype_string_to_value()
%
% Given normalization-type name "normtypestr", returns corresponding
% normalization-type value (or NULL if "normtypestr" is not a valid
% normalization type)
%
{
    variable normtypestr;

    switch (_NARGS)
      { case 1 : normtypestr = (); }
      { usage(sprintf("normtypeval = %s(normtypestr)",_function_name())); }

    normtypestr = strup(strtrim(string(normtypestr)));

    switch (normtypestr)
      { case "NONE" : return NORM_NONE; }
      { case "EFLUX" : return NORM_EFLUX; }
      { case "PFLUX" : return NORM_PFLUX; }
      { case "COUNTS" : return NORM_COUNTS; }
      { case "COUNT_RATE" : return NORM_COUNT_RATE; }
      { return NULL; }
}


public define fakeit_norm()
%
% Uses FAKEIT to generate a simulated dataset with user-specified
% normalization.  The input parameters are as follows:
%
%   dataset:	Dataset number for simulated data
%   exposure:	Exposure time for simulated data
%   backscale:  Background scale factor for simulated data
%   normtype:	Normalization type (integer or string)
%   normval:    Normalization value (i.e. energy flux, photon flux, counts,
%		counts, or count rate over specified energy range
%   normparam:	Name of model parameter to use as normalization
%   normrange:  2-element array containing lower and upper integration bounds
%		for calculating normalization (or NULL to integrate over full
%		range of energy grid)
%
{
    variable dataset = NULL;
    variable exposure;
    variable backscale;
    variable normtype = NULL;
    variable normval = NULL;
    variable normparam = NULL;
    variable normrange = NULL;

    % Get arguments

    switch (_NARGS)
      { case 2 : (exposure,backscale) = (); }
      { case 3 : (dataset,exposure,backscale) = (); }
      { case 6 : (exposure,backscale,normtype,normval,normparam,normrange) = (); }
      {
        case 7 :
          (dataset,exposure,backscale,normtype,normval,normparam,normrange) = ();
      }
      {
        usage(_function_name() + "([dataset,]exposure," +
	  "backscale[,normtype,normval,normparam,normrange])");
      }


    %
    % Validate arguments
    %

    if (dataset == NULL)
        dataset = 1;
    else if (typeof(dataset) != Integer_Type)
	do_error("dataset number must be an integer");

    if (exposure != NULL)  exposure = typecast(exposure,Double_Type);
    if (backscale != NULL)  backscale = typecast(backscale,Double_Type);

    if (normtype != NULL) {
	if (typeof(normtype) == String_Type) {
	    % "normtype" is a string, so convert it to appropriate value
	    normtype = normtype_string_to_value(normtype);
	    if (normtype == NULL)
	        do_error("invalid normalization type");
        } else if (typeof(normtype) == Integer_Type) {
	    % "normtype" is an integer, so make sure it's valid
	    !if (is_valid_normtype(normtype))
		do_error("invalid normalization type");
	} else {
	    % "normtype" isn't NULL, a string, or an integer, so die
	    do_error("normalization type must be a string or an integer");
	}


	% If normalization was requested, validate other normalization
	% parameters

	if (normtype == NORM_NONE) {
	    normtype = NULL;
	} else {
	    % Validate normval
	    normval = typecast(normval,Double_Type);

	    % Validate normparam

	    if (typeof(normparam) != String_Type)
		do_error("normalization parameter must be a string");

	    variable normparam_struct = get_par(normparam);

	    if (typeof(normparam_struct) != Struct_Type)
		do_error(sprintf("invalid model parameter: '%s'",normparam));
	    else if (typeof(normparam_struct.value) == String_Type)
		do_error(sprintf("value of parameter '%s' is not numeric"
		  ,normparam));

	    % Validate normrange

	    if (normrange != NULL) {
		if (length(normrange) != 2)
		    do_error("invalid normalization range");
		normrange = typecast(normrange,Double_Type);
	    }
	}
    }


    %
    % Check that source model and evaluation grid are defined
    %

    if (NULL == get_source_expr(dataset))
    	do_error("no source model defined for dataset " + string(dataset));

    % If neither a dataspace nor an instrument model exists, there's no
    % evaluation grid, and we can't run fakeit
    if ((get_axes(dataset) == NULL) and
      ((get_inst_expr(dataset) == NULL) or (get_inst_expr(dataset) == "")))
	do_error("no grid defined for dataset " + string(dataset));


    %
    % Set FAKEIT parameters
    %

    do_message(3,sprintf("Running FAKEIT for dataset %d",dataset));

    % Set exposure time
    if (exposure == NULL) {
	do_message(4,"Using default FAKEIT exposure time");
    } else {
	do_message(4,sprintf("Setting FAKEIT exposure time to %g",
	  exposure));
        !if (set_fakeit_time(dataset,exposure))
	    do_error("unable to set FAKEIT exposure time");
    }

    % Set backscale
    if (backscale == NULL) {
	do_message(4,"Using default FAKEIT backscale");
    } else {
	do_message(4,sprintf("Setting FAKEIT backscale to %g",backscale));
        !if (set_fakeit_backscale(dataset,backscale))
	    do_error("unable to set FAKEIT backscale");
    }


    %
    % Run FAKEIT
    %

    do_message(4,"Running FAKEIT");
    !if (run_fakeit(dataset))  do_error("unable to run FAKEIT");

    % If no normalization was requested, quit here
    if (normtype == NULL)  return;


    %
    % Calculate normalization value
    %

    do_message(3,sprintf("Re-running FAKEIT to normalize dataset %d",
      dataset));

    if (normrange == NULL) {
	% No normalization range given, so use default range
        do_message(4,"Calculating normalization over default range");
    } else {
	%
	% Normalization range given, so use it
	%

	% Get lower/upper boundaries of dataspace

	variable axis = get_axes(dataset);
	variable axis_length;
	variable axis_min;
	variable axis_max;

	if (axis.mid != NULL) {
	    axis_length = length(axis.mid);
	    axis_min = axis.mid[0];
	    axis_max = axis.mid[length-1];
	} else {
	    axis_length = length(axis.lo);
	    axis_min = (axis.lo[0] + axis.hi[0]) / 2.0;
	    axis_max = (axis.lo[axis_length-1] + axis.hi[axis_length-1]) / 2.0;
	}

	%% THESE PROBABLY SHOULD BE COMPARED TO WITHIN A TOLERANCE

	if (normrange[0] < axis_min) {
	    do_message(4,sprintf("Lower bound for normalization (%g) is below " +
	      "lower bound for dataset (%g);\n  using %g as lower bound",
	      normrange[0],axis_min,axis_min));

	    normrange[0] = axis_min;
	}

	if (normrange[1] == 0.0) {
	    do_message(4,sprintf("Upper bound for normalization is zero; " +
	      "interpreting this as request to use upper bound of dataset " +
	      "(%g)",axis_max));

	    normrange[1] = axis_max;
	} else if (normrange[1] > axis_max) {
	    do_message(4,sprintf("Upper bound for normalization (%g) is above " +
	      "upper bound for dataset (%g);\n  using %g as upper bound",
	      normrange[1],axis_max,axis_max));

	    normrange[1] = axis_max;
	}


	% Check that lower bound is smaller than upper bound
	if (normrange[0] > normrange[1])
	    do_error("lower bound for normalization is greater than " +
	      "upper bound");

        do_message(4,sprintf("Calculating normalization over range " +
	  "(%g:%g)",normrange[0],normrange[1]));
    }

    % Calculated normalization value
    variable norm_calc;

    switch (normtype)
      { case NORM_EFLUX :
          % Get energy flux (in ergs/cm^2/s) over specified range

          norm_calc = get_eflux(dataset,normrange);
	  if (norm_calc == NULL)
	      do_error("unable to normalize by energy flux");
	  do_message(5,sprintf("  Energy flux over normalization range " +
	    "= %g",norm_calc.value));
      }
      { case NORM_PFLUX :
	  % Get photon flux (in photons/cm^2/s) over specified range

	  norm_calc = get_pflux(dataset,normrange);
	  if (norm_calc == NULL)
	      do_error("unable to normalize by photon flux");
	  do_message(5,sprintf("  Photon flux over normalization range " +
	    "= %g",norm_calc.value));
      }
      { case NORM_COUNTS :
          % Get summation of source counts over specified range

	  norm_calc = get_net_counts_sum(dataset,normrange);
	  if (norm_calc == NULL)
	      do_error("unable to normalize by source counts");
	  do_message(5,sprintf("  Summation of source counts over " +
	    "normalization range = %g",norm_calc.value));
      }
      { case NORM_COUNT_RATE :
	  % Calculate source count rate over specified range

	  norm_calc = get_net_counts_sum(dataset,normrange);
	  if (norm_calc == NULL)
	      do_error("unable to normalize by source count rate");
	  norm_calc.value /= get_exptime(dataset);
	  do_message(5,sprintf("  Source count rate over " +
	    "normalization range = %g",norm_calc.value));
      }


    %
    % Normalize simulated data
    %

    % Store ratio of user-specified normalization value to calculated value
    do_message(5,sprintf("  Normalization constant = %g / %g = %g",
      normval,norm_calc.value,normval/norm_calc.value));
    norm_calc = normval/norm_calc.value;

    % Multiply value of normalization parameter by calculated normalization
    % constant
    do_message(4,sprintf("Multiplying value of parameter '%s' by " +
      "normalization constant %g",normparam,norm_calc));
    normparam_struct.value = norm_calc * normparam_struct.value;
    !if (set_par(normparam_struct))
	do_error(sprintf("unable to set value of parameter '%s'"));

    % Re-run FAKEIT
    do_message(4,"Running FAKEIT with calculated normalization");
    !if (run_fakeit(dataset))
	do_error("unable to run FAKEIT with calculated normalization");

    % Done!
}


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Fit-related functions
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


% Confidence-interval estimation methods (defining them as functions makes
% them act as constants, i.e. users can use but not change their values)
static define PARAMEST_NONE() { return 0; }
static define PARAMEST_UNC() { return 1; }
static define PARAMEST_COV() { return 2; }
static define PARAMEST_PROJ() { return 3; }

% Minimum and maximum confidence-interval estimation method values
private define PARAMEST_METHOD_MIN() { return 0; }
private define PARAMEST_METHOD_MAX() { return 3; }


private define is_valid_paramest_method(method)
%
% Returns true if integer "method" corresponds to a valid confidence-interval
% estimation method.
%
{
    return ((method >= PARAMEST_METHOD_MIN) and (method <= PARAMEST_METHOD_MAX));
}


static define paramest_method_string_to_value()
%
% Given confidence-interval estimation method name "methodstr", returns
% corresponding value (or NULL if "methodstr" is not a valid method name).
%
{
    variable methodstr;

    switch (_NARGS)
      { case 1 : methodstr = (); }
      { usage(sprintf("methodval = %s(methodstr)",_function_name())); }

    methodstr = strup(strtrim(string(methodstr)));

    switch (methodstr)
      { case "NONE" : return PARAMEST_NONE; }
      { case "UNC" : return PARAMEST_UNC; }
      { case "COV" : return PARAMEST_COV; }
      { case "PROJ" : return PARAMEST_PROJ; }
      { return NULL; }
}


public define fit_paramest()
%
% Runs fit and specified confidence-interval estimation method and returns
% goodness-of-fit and confidence-interval structures.  Also calls GOODNESS
% command after fit so that goodness-of-fit info is displayed to screen.
%
% Note that this is just a convenience function that provides no new
% functionality.
%
{
    variable dataset = NULL;
    variable method;

    % Get arguments

    switch (_NARGS)
      { case 1 : method = (); }
      { case 2 : (dataset,method) = (); }
      {
        usage(sprintf("(good,paramest) = %s([dataset,]paramest_method)",
	  _function_name()));
      }


    %
    % Validate arguments
    %

    if (dataset == NULL)
        dataset = 1;
    else if (typeof(dataset) != Integer_Type)
	do_error("dataset number must be an integer");

    if (method == NULL) {
	method = PARAMEST_NONE;
    } else if (typeof(method) == String_Type) {
	% "method" is a string, so convert it to appropriate value
	method = paramest_method_string_to_value(method);
	if (method == NULL)
	    do_error("invalid parameter-estimation method");
    } else if (typeof(method) == Integer_Type) {
	% "method" is an integer, so make sure it's valid
	!if (is_valid_paramest_method(method))
	    do_error("invalid parameter-estimation method");
    } else {
	% "method" is neither a string nor an integer, so die
        do_error("parameter-estimation method must be a string or an integer");
    }


    %
    % Run fit and display goodness-of-fit info
    %

    do_message(3,sprintf("Running fit for dataset %d\n",dataset));
    variable good = run_fit(dataset);

    if (good == NULL)
	do_error("unable to fit dataset " + string(dataset));

    % Print goodness info (no reason for this to fail, so don't need to
    % check return value)
    () = sherpa_eval(sprintf("GOODNESS %d",dataset));

    % If no confidence estimation was requested, stop here
    if (method == PARAMEST_NONE)  return (good,NULL);


    %
    % Calculate confidence intervals (if requested)
    %

    % Output of confidence-interval estimation routine
    variable paramest;

    switch (method)
      { case PARAMEST_UNC :
	  do_message(3,"Estimating confidence intervals with UNCERTAINTY");
          paramest = run_unc();
      }
      { case PARAMEST_COV :
	  do_message(3,"Estimating confidence intervals with COVARIANCE");
          paramest = run_cov();
      }
      { case PARAMEST_PROJ :
	  do_message(3,"Estimating confidence intervals with PROJECTION");
          paramest = run_proj();
      }

    if (paramest == NULL)
	do_error("unable to obtain confidence intervals for dataset "
	  + string(dataset));

    % Return goodness-of-fit and confidence-interval structures
    return (good,paramest);
}


