%------------------------------------------------------------------------
%
% filename: spectrum.sl
% Version: 2.1 (11 July 2004)
% purpose: to calculate spectral weights for input to 
%          mkinstmap; written in S-Lang
%
%
% History:
% original version: March 2001
% patch 1: 3 June 2002 - added two lines to prevent "Array index out
%  of allowed range" error.
% patch 2: 11 July 2004 - removed sum() function as it is now part of
%  the S-Lang interpreter provided by Sherpa (in CIAO 3.1)
%  
%
% send comments via CXC Helpdesk
% http://cxc.harvard.edu/helpdesk/
%
%------------------------------------------------------------------------

% prior to S-Lang 1.4.7 the sum() function was not part of the
% S-Lang Run-Time Library. It was added in v1.4.7 but an extra call
% needs to be made whensetting up the S-Lang interpreter to make this
% function available. As of CIAO 3.1 the S-Lang interpreter embedded in
% Sherpa and ChIPS (and also slsh) includes the sum function.
%
% We embed the old function in a conditional statement as a
% precaution.
%

#ifnexists sum
% sum() computes the sum of elements of an array.

define sum (x)
{
  variable num = length(x);
  variable total, i;
  
  total = 0;
  _for (0, num-1, 1)
    {
       i = ();
       total += x[i];
    }
    
  return total;
}

#endif

% integrate_binned() assumes the following:
% f = struct {val, elo, ehi}
%    f.val[k] = \int_f.elo[k]^f.ehi[k] g(e) de
%    f.elo[k] < f.ehi[k]
%    f.elo[k] < f.elo[k+1]
% 
% Integration limits:
%    a < b
%    
% Usage:  result = integrate_binned (f, a, b);
%
%  result is the integral of 'f' from a -> b.
%  
private define integrate_binned (f, a, b)
{
   variable i, result;
   
   i = where ((a < f.elo) and (f.ehi < b));
   if (length(i) == 0)                       % to prevent error
     return 0.0;                             % to prevent error
   result = sum(f.val[i]);

   % Include the endpoint corrections:
   
   variable k, ni, nf, frctn;

   k = i[0];
   if (k > 0)
     {
	frctn = (f.ehi[k-1] - a) / (f.ehi[k-1] - f.elo[k-1]);
	result += f.val[k-1] * frctn;
     }
   
   nf = length(f.val);
   ni = length(i);
   
   k = i[ni-1];
   if (k < nf - 1)
     {
	frctn = (b - f.elo[k+1]) / (f.ehi[k+1] - f.elo[k+1]);
	result += f.val[k+1] * frctn;
     }
   
   return result;
}

typedef struct
{
   elo, ehi
}
Energy_Band_Type;

private define init_bands (elo, ehi)
{
   variable dest = @Energy_Band_Type;
   dest.elo = @elo;
   dest.ehi = @ehi;
   
   return dest;
}

private variable Standard_Bands = Assoc_Type [Energy_Band_Type];

% The band positions listed here  are for illustration only -- 
% the real values should be chosen using the effective area curves.
% 
Standard_Bands["ACIS"] = init_bands ([0.5, 1.0, 1.8, 2.0, 4.0, 6.0,  8.0],
				     [1.0, 1.8, 2.0, 4.0, 6.0, 8.0, 10.0]);
Standard_Bands["HRC"]  = init_bands ([0.5, 2.0, 4.0,  8.0],
		 		     [2.0, 4.0, 8.0, 10.0]);

public define get_standard_bands ()
{
   variable detector;
   
   if (_NARGS != 1)
     {
	message ("Usage:  bands = get_standard_bands (\"detector\")");
	return;
     }
   
   detector = ();
   
   if (0 != assoc_key_exists (Standard_Bands, detector))
     return Standard_Bands[detector];

   vmessage ("Unrecognized detector name:  %s", detector);
}

% calc_instmap_weights assumes that
%       f = struct {val, elo, ehi}
%       bands = struct {elo, ehi}
%       
% Usage:  weights = calc_instmap_weights (f, bands);
% 
public define calc_instmap_weights ()
{
   if (_NARGS != 2)
     {
	message ("Usage:  weights = calc_instmap_weights (spectrum, bands)");
	return;
     }
   
   variable f, bands;
   (f, bands) = ();
   
   variable i, n, weights, norm;

   n = length (bands.elo);
   weights = Float_Type[n];
   
   _for (0, n-1, 1)
     {
	i = ();
	weights[i] = integrate_binned (f, bands.elo[i], bands.ehi[i]);
     }
   
   norm = sum(weights);
   
   if (norm > 0.0)
     return weights / norm;

   vmessage ("Invalid weights:  sum = %f", norm);
   return weights;
}

public define write_weights_file ()
{
   variable filename, weights, bands;
   
   if (_NARGS != 3)
     {
	message ("Usage:  write_weights_file (\"filename\", weights, bands)");
	return;
     }
   
   (filename, weights, bands) = ();
   variable fp = fopen (filename, "w");
   
   if (NULL == fp)
     {
	vmessage ("Failed opening file %s for writing", filename);
	return;
     }
   
   variable i,n = length(weights);

   _for (0, n-1, 1)
     {
	i = ();
	() = fprintf (fp, "%10.3e %10.3e\n", 
		      0.5 * (bands.elo[i] + bands.ehi[i]), weights[i]);
     }
   
   () = fclose (fp);
}

private define load_model_spectrum (file)
{
   variable s = readascii(file);     % read the sherpa model file
   variable n = s._nrows;

   variable f = struct {val, elo, ehi};

   % assuming col2 is the bin-integral:
   
   f.elo = s.col1[[0:n-2]];   % bin low edge
   f.ehi = s.col1[[1:n-1]];   % bin high edge
   f.val = s.col2[[1:n-1]];   % bin integral
   
   return f;
}

public define runtest (file)
{
   variable f, b, w;
   f = load_model_spectrum (file);
   b = get_standard_bands ("ACIS");
   w = calc_instmap_weights (f, b);
   write_weights_file ("weights.txt", w, b);
}



