%; Time-stamp: <03/06/25 12:12:11 bish> 
%; File: tg_scale_reg.sl
%; Author: K. Tibbetts, D. Huenemoerder
%; Original version: 2003.06
%; Updates: 30 July 2003 (ecg) - rearranged usage messages
%;          04 Aug 2003  (ecg) - version matched to tg_scale_reg
%;====================================================================
% Version: 1.4 (12 December 2005)
%
% The script generates ASCII region file containing a "grating
% ruler".  It is used to measure grating dispersion distances 
% on any Chandra grating observation (ACIS/HRC, HETG/LETG).
%

%% Change this, if you don't like the colors
static variable color=["white","green","red","blue","cyan"];

static variable KEV_ANG=12.39854;          %% ang * keV

%%%%% Wavelength scales
static variable waves_heg=[0.5,1.0,2.0,4.0,6.0,8.0,10.0];
static variable waves_meg=waves_heg * 2.0;
static variable waves_leg=[10.0, 20.0, 40.0, 60.0, 80.0];

static variable LEN=10;                    %% Length of line to go along w/label

%%%%% Cal. data, data-struct setup; data are from telD1999-07-23geomN0004.fits
static variable ROWLAND=8632.48;            %% mm

%% Periods in angstroms.  LEG period is adjusted to compensate for different
%%	HETG and LETG ROWLAND spacings
static variable HEG_PERIOD=2000.810;       
static variable MEG_PERIOD=4001.410;       
static variable LEG_PERIOD=9912.160*8632.48/8637.0;  

static variable ACIS_SCALE=0.023987;          %% mm/pixel
static variable HRC_SCALE =0.0064294;      %% mm/pixel


static variable CalData= struct {
  rowland, period, xscale
};

%%%%% End, calibration data, data-struct setup

%%%% Wavelength scale functions
%%% Waves.waves will be an associative array --> each
%%% array will hold the wavelengths for heg, meg or leg
static variable Waves= struct {
  waves
};

%%%%%%% Define the data structures
!if (is_defined ("Box_Type")) {

typedef struct
{
   xcenter, ycenter, xlength, ylength

} Box_Type;

typedef struct {

   xcenter, ycenter, xlength, ylength, rot_angle

} RotBox_Type;

typedef struct {

   xcenter, ycenter, radius

} Circle_Type;

%% The grating variable is needed in the Shape_Type
%% structure in order to be able to apply the appropriate
%% scale to the shape. The choices are "heg", "meg" & "leg"
typedef struct {

   grating, circle, box, rotbox

} Shape_Type;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
%%% This structure holds the shape data
%%% for each source in the file. The variable
%%% "source" below will actually be an array
%%% of type Shape_Type --> it will be of length
%%% NCOMPONENT, the number of components for
%%% each source.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
typedef struct {
 
  ncomps, source

} Source_Type;

}   %%% End, structure definitions
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% print_usage prints the allowed calling sequences for tg_scale_reg.
%   It is invoked when improper syntax is used

define print_usage ()
{

    variable use_string1, use_string2, use_string3;
    variable use_string4, use_string5;
    variable opt_str1, opt_str2, opt_str3;

    use_string1= "Run tg_scale_reg with a FITS input file:";
    use_string2= "   ()=tg_scale_reg (\"infile\" ,\"outfile\")";

    use_string3= "or without an input file:";
    use_string4= "   ()=tg_scale_reg (\"x=<xval>\", \"y=<yval>\", \"rot_ang=<val>\",";
    use_string5= "   \"grat=<val>\", \"inst=<val>\", \"outfile=<val>\")";
 
    opt_str1= "\tIn either mode, the \"scale\" and \"values\" parameters are optional.\n\n";
    opt_str1+="\t   [,\"scale=<scale>\"][,\"values=v1,v2,v3,...\"])\n\n";
    opt_str1+= "\tThe \"scale\" may be \"kev\" or \"angstrom\" (default).\n\n";

    opt_str2= "\tThe location of the tick marks is given as a comma-separated list\n";

    opt_str2+="\tin the \"values\" parameter and should be in the same units as\n";
    opt_str2+="\t\"scale\".  If a FITS file is used and the grating is HETG, the\n";
    opt_str2+="\tsame values will be be used for both the HEG and MEG arms.\n";

    opt_str3= "\n\tNote: all quotes shown above are REQUIRED.\n\n";

    ()=fprintf (stderr, "\n\t%s\n\n\t%s\n", use_string1, use_string2);
    ()=fprintf (stderr, "\n\t%s\n", use_string3);
    ()=fprintf (stderr, "\n\t%s\n\t%s\n\n", use_string4, use_string5);
    ()=fprintf (stderr, "\twhere:\n");
    ()=fprintf (stderr, "\n\t\"grat=<val>\" -> leg, hetg, heg, or meg\n");
    ()=fprintf (stderr, "\t\"inst=<val>\" -> acis or hrc\n");
    ()=fprintf (stderr, "\n\tOptional Parameters:\n\n");
    ()=fprintf (stderr, "%s", opt_str1);
    ()=fprintf (stderr, "%s", opt_str2);
    ()=fprintf (stderr, "%s", opt_str3);

}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% The following functions determine the method used to invoke
% tg_scale_reg and verify the parameters are of the proper form
define check_x_y_rot (value)
{
    variable r;

    if (1 != sscanf (value, "%f", &r))
     {
       ()=fprintf (stderr, "\n\tError: Cannot parse a float for\n");
       ()=fprintf (stderr, "\tx/y/rot_ang (value=\"%s\")\n\n", value);
       return -1;
     }

    return 0;
}

define check_grat (value)
{

    if ( (value == "heg") or (value == "meg") or (value == "hetg")
	  or (value=="letg") or (value=="leg"))
     return 0;
    
    ()=fprintf (stderr, "\n\tIllegal value for grating!\n");
    ()=fprintf (stderr, "\tvalue=%s\n", value);
    ()=fprintf (stderr, "\tChoices are \"letg\",\"hetg\", \"heg\" and \"meg\"\n\n");

    return -1;
}

define check_inst (value)
{

    if ((value == "acis") or (value=="hrc"))
     return 0;

    ()=fprintf (stderr, "\n\tIllegal value for instrument!\n");
    ()=fprintf (stderr, "\tvalue=%s\n", value);
    ()=fprintf (stderr, "\tChoices are \"acis\" or \"hrc\"\n\n");
    return -1;

}

define NEWparse_args (nargs)
{

%%%%+
%%  There are two ways in which this program may be used:
%%  The user can specify an input file, from which all
%%  necessary info. will be extracted. Alternatively, the
%%  user may specify the grating, instrument, x & y values
%%  (indicating the source pos.) and the roll angle.
%%  In the latter case, the command-line args. for these
%%  values are held in the variable no_infile_data.
%%%%-

    variable infile=NULL, outfile=NULL,vallist=NULL,values=NULL;
    variable scale,nvals,i;
    variable no_infile_data= struct {grat, inst, x, y, rot_ang};
    variable this_arg,this_name,this_val;
    scale="LAMBDA";

    if (nargs < 2)
     {
        print_usage();
        return -1, NULL, NULL, NULL, NULL,NULL;
     }

    if (nargs > 4) loop (nargs)
     {

       %% I do no error checking re: the input type of the
       %% arguments. (All arguments must be strings!) I will
       %% merely convert any non-string args. into strings

       this_arg=string (());

       %% Look for options. Options are specified 
       %% via "option-name=value"
       this_name= extract_element (this_arg, 0, '=');
       this_val = extract_element (this_arg, 1, '=');

       if (this_val == NULL)
        {
	    print_usage();
	    return,-1,NULL,NULL,NULL,NULL,NULL;
        }
       else {
         if (strlow (this_name) == "scale")
          {
            scale=strlow (this_val);
            if (scale != "kev" and scale != "angstrom")
             {  
                ()=fprintf (stderr, "\n\tIllegal choice for SCALE!\n");
                ()=fprintf (stderr, "\tChoices are \"kev\" or \"angstrom\"\n");
                ()=fprintf (stderr, "\t(The case does not matter.)\n\n");
                return -1, NULL, NULL, NULL, NULL,NULL;
             }
          }
         else if (strlow (this_name) == "x")
          {
            if (-1 == check_x_y_rot (this_val))
             return -1, NULL, NULL, NULL, NULL,NULL; 
            no_infile_data.x=strlow (this_val); 
          }
         else if (strlow (this_name) == "y")
          {
            if (-1 == check_x_y_rot (this_val))
             return -1, NULL, NULL, NULL, NULL,NULL; 
            no_infile_data.y=strlow (this_val); 
          }
         else if (strlow (this_name) == "rot_ang")
          {
            if (-1 == check_x_y_rot (this_val))
             return -1, NULL, NULL, NULL, NULL,NULL; 
            no_infile_data.rot_ang=strlow (this_val); 
          }
         else if (strlow (this_name) == "grat")
          {
            if (-1 == check_grat (strlow (this_val)))
             return -1, NULL, NULL, NULL, NULL,NULL; 
	    if (this_val == "letg") this_val="leg";
            no_infile_data.grat=strlow (this_val); 
          }
         else if (strlow (this_name) == "inst")
          {
            if (-1 == check_inst (strlow (this_val)))
             return -1, NULL, NULL, NULL, NULL,NULL; 
            no_infile_data.inst=strlow (this_val); 
          }
         else if (strlow (this_name) == "outfile")
          {
            outfile=this_val;
          }
	 else if (strlow(this_name) == "values")
	  {
	    vallist=strchop(strcompress(this_val,", "),',',0);
	    nvals=length(vallist);
	    values=Double_Type[nvals];
	    for (i=0;i<nvals;i++) values[i]=atof(vallist[i]);
          }
         else
          {
           ()=fprintf (stderr, "\n\tERROR: Unrecognized option.\n\n");
           ()=fprintf (stderr, "\tPossible options:\n");
           ()=fprintf (stderr, "\t\"scale=<angstrom/kev>\"\n");
           ()=fprintf (stderr, "\t\"values=<comma list>\"\n");
           ()=fprintf (stderr, "\tThe quotes are necessary!\n\n");
           return -1, NULL, NULL, NULL, NULL,NULL;
          }
        }
     }
     else
      {
        loop(nargs -2)
	{
	  this_arg=string(());
	  this_name= extract_element (this_arg, 0, '=');
          this_val = extract_element (this_arg, 1, '=');
	  if (this_val == NULL)
           {
	    print_usage();
	    return,-1,NULL,NULL,NULL,NULL,NULL;
           }
	  if (strlow (this_name) == "scale")
           {
            scale=strlow (this_val);
            if (scale != "kev" and scale != "angstrom")
              {  
                ()=fprintf (stderr, "\n\tIllegal choice for SCALE!\n");
                ()=fprintf (stderr, "\tChoices are \"kev\" or \"angstrom\"\n");
                ()=fprintf (stderr, "\t(The case does not matter.)\n\n");
                return -1, NULL, NULL, NULL, NULL,NULL;
              }
            }
	 else if (strlow (this_name) == "values")
          {
	    vallist=strchop(strcompress(this_val,", "),',',0);
	    nvals=length(vallist);
	    values=Double_Type[nvals];
	    for (i=0;i<nvals;i++) values[i]=atof(vallist[i]);
          }
         else
          {
           ()=fprintf (stderr, "\n\tERROR: Unrecognized option.\n\n");
           ()=fprintf (stderr, "\tPossible options:\n");
           ()=fprintf (stderr, "\t\"scale=<angstrom/kev>\"\n");
           ()=fprintf (stderr, "\t\"values=<comma list>\"\n");
           ()=fprintf (stderr, "\tThe quotes are necessary!\n\n");
           return -1, NULL, NULL, NULL, NULL,NULL;
          }
	}
       outfile=string(());
       infile=string(());
     }

%% Now check to see if the user specified either an input file
%%  of the set of inputs required for the 2nd option (ie. x, y,
%%  rot_ang, grat, inst)

    variable error_in_pars;
    error_in_pars=0;
    if (infile == NULL)
     {
        if ((no_infile_data.inst == NULL) or (no_infile_data.grat == NULL)
             or (no_infile_data.x == NULL) or (no_infile_data.y == NULL)
             or (no_infile_data.rot_ang == NULL) )
          error_in_pars=1;
	else 
	   {
	   variable ra=(360.0-atof(no_infile_data.rot_ang)) mod 180;
	   if (no_infile_data.grat == "meg")
	    no_infile_data.rot_ang=string(ra - 4.755);
	   else if (no_infile_data.grat=="heg")
	    no_infile_data.rot_ang=string(ra + 5.205);
	   else if (no_infile_data.grat=="hetg")
	    no_infile_data.rot_ang=[string(ra+5.205),string(ra-4.755)];
	   else no_infile_data.rot_ang=string(ra-0.07);
	   } 
     }
    else
     no_infile_data=NULL;
 
    if (error_in_pars)  
     {
        ()= fprintf (stderr, "\n\tError in parameters:\n");
        ()= fprintf (stderr, "\tAn infile was not given, so values\n");
        ()= fprintf (stderr, "\tbe given for x, y, rot_ang,\n");
        ()= fprintf (stderr, "\tgrat, inst & outfile\n\n");
        print_usage ();
        return -1, NULL, NULL, NULL, NULL,NULL;
     }
    if ((scale == "kev") and (vallist != NULL))
	values=KEV_ANG/values;
     
    return 0, infile,outfile, scale, values, no_infile_data;

}
%%%%%%  End, parameter verification and parsing structures
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

define fill_cal_data_struct (instrument)
{

    CalData.rowland=ROWLAND;   

    CalData.period= Assoc_Type [Double_Type];
    CalData.period ["heg"]=HEG_PERIOD;
    CalData.period ["meg"]=MEG_PERIOD;
    CalData.period ["leg"]=LEG_PERIOD;

    if (instrument=="acis")  CalData.xscale=ACIS_SCALE;
      else CalData.xscale=HRC_SCALE;

    if (instrument=="hrc") LEN=30;
   
}

define fill_waves(values)
{
  Waves.waves= Assoc_Type [Array_Type];
  if (values == NULL)
     {
      Waves.waves ["heg"]=waves_heg;
      Waves.waves ["meg"]=waves_meg;
      Waves.waves ["leg"]=waves_leg;
     }
  else
     {
      Waves.waves["heg"]=values;
      Waves.waves["meg"]=values;
      Waves.waves["leg"]=values;
     }
}


%%%% End, Wavelength structure access functions
 

%%% Misc. functions

define extract_key_name (line)
{
    
    variable name, nchar;

    if (strlen (line) < 8)
     nchar= strlen (line);
    else
     nchar=8;    %%% 8 is the max. length of a keyword

    name=substr (line, 1, nchar);

    if (typeof (name) != Null_Type)
     name= str_delete_chars (name, " ");

    return name;

}
  
%%% For this function I make the assumption that the 
%%% keyword+value+comment in the file are in the form 
%%% KEYWORD = ........         / comment . .. .
%%%   -- or --
%%% KEYWORD = ........         
define get_key_value (line, key)
{
   
    variable pos, len;
    variable start=10;   %%% Corresponds to the fact that the value
                         %%% in a fits header record begins at col. 10

    pos= is_substr (line, "/");

    if (pos == 0)
     len=strlen(line)-start+1;
    else
     len=pos-start;

    variable value;
    value=substr (line, start, len);

    if (typeof (value) != Null_Type)
     {
       value= str_delete_chars (value, "'");
       value= str_delete_chars (value, " ");
     }

    return value;
 
}

%% list is an array of String_Type's 
define find_key_in_list (list, key_to_find)
{
 
    
    foreach (list)
     {
        variable this_line=();
        if ( extract_key_name (this_line) == key_to_find)
         return get_key_value (this_line, key_to_find);
     }
     
    return NULL;
 
}

define check_for_columns (s)
{

    variable cols_to_find= ["x", "y", "r", "rotang", "component",
                            "shape", "grating", "source"];

    variable fields;

    fields=get_struct_field_names (s);

    foreach (cols_to_find)
     {
        variable name;
        name=();
        if (length (where (fields == name)) == 0)
         {
           ()=fprintf (stderr, "\n\tCannot find \"%s\" field in the\n", name);
           ()=fprintf (stderr, "\tinput data structure! Aborting.\n\n");
           return -1;
         }
     }

    return 0;

}

define normalize_comp (component)
{

    variable i, norm;
    
    norm= min (component);

    for (i=0; i<length (component); i++)
     component [i] -= norm;
       
}

define get_circle_data (index, s, ndim, circle)
{

    if (ndim == 2)
     {
       circle.xcenter=s.x[index, 0];  
       circle.ycenter=s.y[index, 0];
     }
    else
     {
       circle.xcenter=s.x[index];  
       circle.ycenter=s.y[index];
     }

    circle.radius= s.r[index, 0]; 

    return 0; 

}

define get_rotbox_data (index, s, ndim, rotbox)
{

    if (ndim==2)
     {
       rotbox.xcenter=s.x[index, 0]; 
       rotbox.ycenter=s.y[index, 0]; 
       rotbox.rot_angle=s.rotang[index, 0];
     }
    else
     {
       rotbox.xcenter=s.x[index]; 
       rotbox.ycenter=s.y[index]; 
       rotbox.rot_angle=s.rotang[index];
     }

    rotbox.xlength=s.r[index, 0];
    rotbox.ylength=s.r[index, 1]; 

    return 0;

}

define get_box_data (index, s, ndim, box)
{

    if (ndim==2)
     {
       box.xcenter=s.x[index, 0]; 
       box.ycenter=s.y[index, 0]; 
     }
    else
     {
       box.xcenter=s.x[index]; 
       box.ycenter=s.y[index]; 
     }

    box.xlength=s.r[index, 0];
    box.ylength=s.r[index, 1]; 

    return 0;

}


define get_one_source_region (src, raw_data,minsource)
{

    variable this_src;
    variable ncomps, grating;

    %%%%%+ 
    %%% Initialize the data structure for the source
    %%% The structure is actually an array of structures
    %%% of type "Shape_Type", each of which contains the
    %%% data for one of the components of the region.
    %%%%%- 
    grating=raw_data.grating [src];

    if (grating == "letg")
     grating="leg";

    %%% We will need to know what grating -- heg, meg, or leg -- that
    %%% the particular shape corresponds to. This is set by the grat
    %%% variable: If the grating column in the region extesion is
    %%% "hetg", the number of components to the region is 3: 0 (zero
    %%% order, 1 (heg) and 2 (meg), etc. (See e.g. ahelp dmrega2fits)
    variable grat;
    if (grating == "hetg")
     {
       grat = ["zero-order", "heg", "meg"];
       ncomps=3;
     }
    else if (grating == "heg" or grating == "meg" or grating == "leg")
     {
       grat = ["zero-order", grating];
       ncomps=2;
     }
    else
     {
       ()=fprintf (stderr, "\n\tError: Cannot handle grating=\"%s\"\n", grating);
       return -1, -1, NULL;  
     }

%%% Normalize the value for COMPONENT (Older files gave component
%%% values starting at 0 while newer ones start at 1. Normalize
%%% them to 0.
    normalize_comp (raw_data.component);
 
    this_src= Shape_Type [ncomps];

    variable i, ndim;

    (,ndim,)=array_info(raw_data.x);
    variable ind,x=where (raw_data.source==src+minsource);
    if (length (x) != ncomps)
        {
        ()=fprintf (stderr, "\n\tCannot locate data in the table!\n");
        ()=fprintf (stderr, "\tCannot find data for source=%d\n\n", src);
        return -1, -1, NULL;
        }  

    for (i=0; i<ncomps; i++)
     { 
        ind=x[i];

        this_src[i].grating= grat[i];
        if (raw_data.shape[i] == "circle")
         {
            this_src[i].circle=@Circle_Type;
            if (-1 == get_circle_data (ind, raw_data, ndim, this_src[i].circle))
             return -1, -1, NULL;
         }
        else if (raw_data.shape[i] == "rotbox")
         {
            this_src[i].rotbox=@RotBox_Type;
            if (-1 == get_rotbox_data (ind, raw_data, ndim, this_src[i].rotbox))
             return -1, -1, NULL;
         }
        else if (raw_data.shape[i] == "box")
         {
            this_src[i].box=@Box_Type;
            if (-1 == get_box_data (ind, raw_data, ndim, this_src[i].box))
             return -1, -1, NULL;
         }
        else
         {
            ()=fprintf (stderr, "\n\tUnrecognized shape-type!\n");
            ()=fprintf (stderr, "\tShape-type=\"%s\"\n", raw_data.shape [i]);
            ()=fprintf (stderr, "\n\tRecognized types are \"circle\",\n");
            ()=fprintf (stderr, "\t\"box\" & \"rotbox\"\n\n");
            return -1, -1, NULL;
         }
            
     } 

    return 0, ncomps, this_src;

}

define read_data (file)
{

    variable region_data;
    variable instrument;
    variable fp,fp2, NSOURCES;

    variable raw_data;   %%% This is the data as read from the file


%%% Check to see if the file exists, read the data if it does
#ifexists readfile
     {
       variable save_case_sens = varmm.caseinsen;
       varmm.caseinsen=1;


       fp=readfile (file+"[REGION]");

       if (fp == NULL)
        {
          ()=fprintf (stderr, "\n\tCannot find the given file!\n"); 
          ()=fprintf (stderr, "\tTried to open \"%s\"\n\n", file);
          return -1, NULL, NULL;
        }

       raw_data=fp;
 
       %%% Now get the instrument

       instrument=find_key_in_list (fp._header, "INSTRUME");
       if (typeof (instrument) == Null_Type)
        {
	   fp2=readfile(file);
	   instrument=find_key_in_list (fp2._header, "INSTRUME");
	   if (typeof (instrument) == Null_Type)
	      {
	         ()=printf(stderr,"\n\tERROR: Could not find INSTRUME keyword in file!\n\n");
                 return -1, NULL, NULL;
	      }
        }

     }
#else
     {
       variable status;

	if (_fits_open_file(&fp,file+"[REGION]","r") != 0)
        {
          ()=fprintf (stderr, "\n\tCannot find the given file!\n");
          ()=fprintf (stderr, "\tTried to open \"%s\"\n\n", file);
          return -1, NULL, NULL;
        }

       status=_fits_read_key (fp, "INSTRUME", &instrument, NULL);
       if (status != 0)
        {
	   ()=_fits_open_file(&fp2,file,"r");
	   status=_fits_read_key (fp2, "INSTRUME", &instrument, NULL);
	   if (status !=0)
	      {
		()=printf(stderr,"\n\tERROR: Could not find INSTRUME keyword in file!\n\n");
           	return -1, NULL, NULL;
	      }
        }
       !if (fp==NULL)
         raw_data=fits_read_table (fp);
     } 
#endif

    instrument= strlow (instrument);

    if (-1 == check_for_columns (raw_data))
      return -1, NULL, NULL;

%%% Determine how many sources
    variable minsource=min(raw_data.source);
    NSOURCES=max (raw_data.source)-minsource+1;
   
    variable word;
    if (NSOURCES > 1)
     word="sources";
    else
     word="source";

    ()=fprintf (stdout, "\n\t** Found %d %s.\n", NSOURCES, word);

%%% Initialize the "master" data structure
    region_data= Source_Type [NSOURCES];

%%% Loop over the bintable data and fill each structure
    variable i, flag;
    for (i=0; i<NSOURCES; i++)
     {
       (flag, region_data[i].ncomps, region_data[i].source)= 
                      get_one_source_region (i, raw_data,minsource);
       if (flag == -1)
        return -1, NULL, NULL;
     }

    instrument= strlow (instrument);

    return 0, instrument, region_data;

}

define build_regions (no_infile_dat)
{
 
   variable region_data, instrument,i;
   variable x, y, rot_ang;
   variable hetg=0;

   ()= sscanf (no_infile_dat.x, "%f", &x); 
   ()= sscanf (no_infile_dat.y, "%f", &y); 

   if (no_infile_dat.grat == "hetg") hetg=1;
   if (hetg) rot_ang=[atof(no_infile_dat.rot_ang[0]),atof(no_infile_dat.rot_ang[1])];
   else  rot_ang=atof(no_infile_dat.rot_ang);
   region_data= Source_Type [1];
   region_data[0].source= Shape_Type [hetg+2];

   region_data[0].source[0].grating= no_infile_dat.grat;
   region_data[0].source[1].grating= no_infile_dat.grat;
   
   region_data[0].source[0].circle=@Circle_Type;    
   region_data[0].source[1].rotbox=@RotBox_Type;    

   region_data[0].source[0].circle.xcenter= x;
   region_data[0].source[0].circle.ycenter= y;
   if (no_infile_dat.inst == "acis") region_data[0].source[0].circle.radius= 8.0;
   else region_data[0].source[0].circle.radius= 30.0;
   if (hetg) 
     {
     region_data[0].source[1].grating="heg";
     region_data[0].source[2].grating="meg";
     region_data[0].source[2].rotbox=@RotBox_Type;
     }

   for (i=0;i<(hetg+1);i++)
     {
     region_data[0].source[i+1].rotbox.xcenter= x;
     region_data[0].source[i+1].rotbox.ycenter= y;
     region_data[0].source[i+1].rotbox.xlength= 0.0;
     region_data[0].source[i+1].rotbox.ylength= 0.0;
     region_data[0].source[i+1].rotbox.rot_angle= rot_ang[i];
     }
     
   instrument=no_infile_dat.inst;
   return 0, instrument, region_data;

}

%%% End, misc. functions

define write_comment (fp, comment)
{
    ()=fprintf (fp, Sprintf ("#%s\n", comment, 1));
}

define write_other (fp, comment)
{
    ()=fprintf (fp, Sprintf ("%s\n", comment, 1));
}

define get_zero (shapes)
{
    
    variable i;

    for (i=0; i< length (shapes); i++)
     {
        if (typeof (shapes[i].circle) != Null_Type)
          return [shapes[i].circle.xcenter, shapes[i].circle.ycenter]; 
     }

    ()=fprintf (stderr, "\n\tError! No zero-order position given in the\n");
    ()=fprintf (stderr, "\        input region extension!\n\n");

    return NULL;

}

define write_circle (fp, circle,srccolor)
{
     variable line;

     line= Sprintf ("circle(%f,%f,%f)\n", circle.xcenter,
         circle.ycenter, circle.radius, srccolor, 4);

     ()=fprintf (fp, line);

}

define write_rotbox (fp, rotbox,srccolor)
{
     variable line;
     line= Sprintf ("rotbox(%f,%f,%f,%f,%f)\n", rotbox.xcenter,
         rotbox.ycenter, rotbox.xlength, rotbox.ylength, rotbox.rot_angle,
         srccolor, 6); 

     ()=fprintf (fp, line);

}

define write_box (fp, box,srccolor)
{
     variable line;

     line= Sprintf ("box %f %f %f %f #color=%s\n", box.xcenter,
         box.ycenter, box.xlength, box.ylength, srccolor, 5);

     ()=fprintf (fp, line);

}

define write_shape (fp, shape,srcid)
{
    variable srccolor=color[srcid mod 5];
    if (typeof (shape.circle) != Null_Type)
     write_circle (fp, shape.circle,srccolor);
    else if (typeof (shape.rotbox) != Null_Type)
     write_rotbox (fp, shape.rotbox,srccolor);
    else
     write_box (fp, shape.box,srccolor);

}

define get_waves (grat)
{

   !if (assoc_key_exists (Waves.waves, grat))
     return NULL;

   return Waves.waves [grat];

}

define get_ticks (grating, scale)
{

     if (scale=="kev")
      return KEV_ANG / get_waves (grating);
     else
      return get_waves (grating);

}






%%%%+
%% The output file name will be given by root + basename + .reg
%% where basename is the name of the file up to the extension.
%% e.g. if the inupt file name is  input.file.fits the output
%% file will be input.file.reg, etc. 
%%%%-




define get_line_coords (wave, zero, xscale, angle, grating) 
{

     variable tg_r, z;
     variable delta, phi;
     variable xm, ym, xp, yp;
     variable xml=Float_Type[2];
     variable xpl=xml,yml=xml,ypl=xml;
     variable cp, cm;
     variable dline,etaline,xiline; 
     tg_r= wave * (CalData.rowland / CalData.period [grating]) / xscale;  

     variable xi, eta, d;
     if (grating == "meg")
      {
         d=-LEN * 8;          %% The 8 is arbitrary; it seems to work ok
         cp=""; cm="";
      }
     else 
      {
         d=LEN * 8;
         cp=""; cm="";
      }
     dline=d/4.;

     %%% Now specify coords. for the text, positive orders
     eta=atan (d/tg_r);
     xi=sqrt (d*d + tg_r*tg_r);
     xp=xi * cos (angle + eta) + zero [0];
     yp=xi * sin (angle + eta) + zero [1];


     %%% Now specify coords. for the text, negative orders
     eta += 3.14159265;
     xm=xi * cos (angle + eta) + zero [0];
     ym=xi * sin (angle + eta) + zero [1];
     

     dline=d/4.0;
     etaline=atan(dline/tg_r);
     xiline=sqrt(dline^2+tg_r^2);
     xpl=[tg_r*cos(angle)+zero[0],xiline*cos(angle+etaline)+zero[0]];
     ypl=[tg_r*sin(angle)+zero[1],xiline*sin(angle+etaline)+zero[1]];
     xml=[tg_r*cos(angle+3.14159)+zero[0],xiline*cos(angle+3.14159+etaline)+zero[0]];
     yml=[tg_r*sin(angle+3.14159)+zero[1],xiline*sin(angle+3.14159+etaline)+zero[1]]; 
      
     return xm, ym, xp, yp, cp, cm,xpl,ypl,xml,yml;

}

define write_ruler (fp, source,  scale,srcid)
{

     variable zero, xscale;
     variable waves;
     variable ticks;        %%% ticks are the 'tick' marks of the 
                            %%% ruler. ie. energies or wavelengths
     zero=get_zero (source);
     if (typeof (zero) == Null_Type)
      return -1; 

     xscale=CalData.xscale;

     variable i, j;
     variable cp, cm;  %% This is either '<' or '>'

     for (i=0; i < length (source); i++)
      {
         variable x0, y0, x1, y1;
         variable xm, ym;           %%% x,y values for the text, neg. order
         variable xp, yp;           %%% x,y values for the text, pos. order
	 variable xpl,ypl,xml,yml;
%         write_comment (fp, Sprintf (".COMPONENT %d : tg_part=%s", i,
%                         source [i].grating, 2));
         write_shape (fp, source [i],srcid);

         if (typeof (source[i].circle) != Null_Type)
          continue;
 
         waves=get_waves (source[i].grating);
         ticks=get_ticks (source[i].grating, scale);

         variable angle;

         if (typeof (source[i].box) != Null_Type)
          angle=0.0;
         else
          angle=(source[i].rotbox.rot_angle) * 3.14159265 / 180.;
             
      
         for (j=0; j<length (waves); j++)
          {
 	    if (j==0)
 		{
		}
             (xm, ym, xp, yp, cp, cm,xpl,ypl,xml,yml) = get_line_coords (waves [j], zero, xscale, 
                                angle, source[i].grating); 

             ()=fprintf (fp, "# text(%f,%f) text={%3.1f}\n", 
                xm, ym, ticks [j], color[srcid mod 5] );
             ()=fprintf (fp, "# text(%f,%f) text={%3.1f}\n",
                xp, yp, ticks [j], cp, color[srcid mod 5] );
	     ()=fprintf (fp, "line(%f,%f,%f,%f) # line = 0 0\n",
		xpl[0],ypl[0],xpl[1],ypl[1],color[srcid mod 5]);
	     ()=fprintf (fp, "line(%f,%f,%f,%f) # line = 0 0\n",
		xml[0],yml[0],xml[1],yml[1],color[srcid mod 5]);
          }
         
      }

     return 0;

}

define create_output_file (outfile, infile, region_data, scale)
{
	
    variable fp;

    fp= fopen (outfile, "w");
    if (typeof (fp) == Null_Type)
     {
        ()=fprintf (stderr, "\n\tError! Could not create the output file!\n");
        ()=fprintf (stderr, "\t       Tried to create \"%s\"\n\n", outfile);
        return -1;
     }

    variable plate_scale=CalData.xscale;
    variable fstring;

    if (infile != NULL)
      fstring= "Region is for input file " + infile;
    else
      fstring= "Region for user-supplied info.";
 
%    write_comment (fp, fstring);
%    write_comment (fp, "#");
%    write_comment (fp, "#Source numbers are as given in the input fits file");
%    write_comment (fp, "#");
    write_other (fp, "global color=white font=\"helvetica 10 normal\" select=1 highlite=1 edit=1 move=1 delete=1 include=1 fixed=0 source");
    write_other (fp, "physical");

    variable i,j;
    for (i=0; i<length (region_data); i++)
     {

%        write_comment (fp, Sprintf (".SOURCE %d", i+1, 1));

        if (-1 == write_ruler (fp, region_data[i].source, scale,i))
          return -1;

     }
 
    ()=fclose (fp);

    return 0;

}


%%%% This is "main ()"
define tg_scale_reg ()
{

    variable infile, outfile, no_infile_dat;
    variable scale,values;
    variable ticks;      %%% Ticks is the scale -- wavelength or energy --
                         %%% that will be written to the region file
    variable flag;

    infile=NULL;

    (flag, infile, outfile,scale, values, no_infile_dat) = NEWparse_args (_NARGS);
    if (flag == -1)
     return -1;

    fill_waves(values);

    variable instrument, region_data;
    if (no_infile_dat == NULL)
      (flag, instrument, region_data) = read_data (infile);
    else
      (flag, instrument, region_data) = build_regions (no_infile_dat);
     
    if (flag == -1)
     return -1;

    if (typeof (instrument) == Null_Type)
     {
        ()=fprintf (stderr, "\n\tERROR: Cannot find \"INSTRUME\"\n");
        ()=fprintf (stderr, "\t       or instrument value=NULL\n\n");
        return -1;
     }

    fill_cal_data_struct (instrument);

    if (-1 == create_output_file (outfile, infile, region_data, scale))
     return -1;


    ()=fprintf (stdout, "\n\tOutput region file \"%s\" has been created\n\n",
                outfile);

    return 0; 

}
